service system/mountroot {
	depends = system/initial
start {
	create_dm_device()
	{
		mknod --mode=600 /dev/lvm c 109 0 >/dev/null 2>&1 || true
		dm_dir="/dev/mapper"
		dm_file="$dm_dir/control"
		if [ ! -f /dev/.devfsd ] 
		then
			major=$(grep "[0-9] misc$" /proc/devices | sed 's/[ ]\+misc//')
			minor=$(grep "[0-9] device-mapper$" /proc/misc | sed 's/[ ]\+device-mapper//')
			test -d $dm_dir || mkdir --mode=755 $dm_dir >/dev/null 2>&1
			test -c $dm_file -o -z "$major" -o -z "$minor" || mknod --mode=600 $dm_file c $major $minor >/dev/null 2>&1
		fi
	}
	
	
	startLVM() {
		if [ -f /sbin/vgchange ]; then
			# try to load module in case that hasn't been done yet
			modprobe lvm >/dev/null 2>&1
			modprobe lvm-mod >/dev/null 2>&1
			modprobe dm-mod >/dev/null 2>&1
			create_dm_device
			/sbin/vgscan --ignorelockingfailure --mknodes >/dev/null 2>&1
			/sbin/vgchange --ignorelockingfailure -a y >/dev/null 2>&1
		fi
	}


	startEVMS() {
		if [ -f /sbin/evms_activate ]; then
			evms_activate 2>&1 >/dev/null 
		fi
	}

	startDM() {
		if [ -f /sbin/dmsetup ]; then
			/sbin/dmsetup mknodes 2>&1 >/dev/null 
		fi
	}





	export PATH="/bin:/usr/bin:/sbin:/usr/sbin"
	retval=0

	startDM
	startLVM
	startEVMS
		
	if [ ! -f /fastboot ] ; then
	    echo "Remounting root filesystem read-only (if necessary) ..."
	    mount -n -o remount,ro / 2>&1 >/dev/null 

		#TODO FORWARD BOOT PARAMS TO SCRIPTS if [ -f /forcefsck ] || get_bootparam "forcefsck" ; then
		if [ -f /forcefsck ] ; then
			echo "Checking root filesystem (full fsck forced)"
			/sbin/fsck -C -a -f /
			# /forcefsck isn't deleted because system/checkfs needs it.
			# it'll be deleted in that script.
			retval=$?
		else
			# TODO THIS TO if ! is_net_fs / ; then
				# Obey the fs_passno setting for / (see fstab(5))
				# - find the / entry
				# - make sure we have 6 fields
				# - see if fs_passno is something other than 0
				AWKLINE=$(awk '($1 ~ /^\// && $2 == "/" && NF == 6 && $6 != 0) { print }' /etc/fstab)
				if [ -n "$AWKLINE" ]
				then
					echo "Checking root filesystem ..."
					/sbin/fsck -C -T -a /
					retval=$?
				else
					echo "Skipping root filesystem check (fstab's passno == 0) ..."
					retval=0
				fi
			#else
			#	echo "Skipping root filesystem check (networked) ..."
			#	retval=0
			#fi
		fi
		if [ ${retval} -eq 0 ] ; then
			echo "Success .."
		elif [ ${retval} -eq 1 ] ; then
			echo "Filesystem repaired ..."
		elif [ ${retval} -eq 2 ] || [ ${retval} -eq 3 ] ; then
			echo "Filesystem repaired, but reboot needed!"
			echo -ne "\a"; sleep 1; echo -ne "\a"; sleep 1
			echo -ne "\a"; sleep 1; echo -ne "\a"; sleep 1
			echo "Rebooting in 2 seconds ..."
			sleep 2
			echo "Rebooting"
			/sbin/ngc -j reboot
		else
			echo "Filesystem couldn't be fixed :("
			/sbin/ngc -j sulogin
			exit ${retval}
		fi
	fi

	# Should we mount root rw ?
	if mount -vf -o remount / 2> /dev/null | awk '{ if ($6 ~ /rw/) exit 0; else exit 1; }'
	then

	    echo "Remounting root filesystem read/write"
	    mount -n -o remount,rw / 2>&1 >/dev/null 
	    if [ "$?" -ne 0 ]
	    then
		echo "Root filesystem could not be mounted read/write :("
		/sbin/ngc -j sulogin
	    fi
	fi

	#
	# Create /etc/mtab
	#
	# Don't create mtab if /etc is readonly
	if ! touch /etc/mtab 2> /dev/null
	then
		echo "Skipping /etc/mtab initialization (ro root?)"
		exit 0
	fi

	# Clear the existing mtab
	> /etc/mtab

	# Add the entry for / to mtab
	mount -f /

	# Don't list root more than once
	awk '$2 != "/" {print}' /proc/mounts >> /etc/mtab
	# Now make sure /etc/mtab have additional info (gid, etc) in there

	for x in $(awk '{ print $2 }' /proc/mounts | sort -u)
	do
		for y in $(awk '{ print $2 }' /etc/fstab)
		do
			if [ "${x}" = "${y}" ]
			then
				mount -f -o remount $x
				continue
			fi
		done
	done

	# Remove stale backups
	rm -f /etc/mtab~ /etc/mtab~~

}
}
