
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"
#include "../../src/initng_handler.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include "../../src/initng_global.h"
#include "../../src/string_tools.h"
#include "../../src/active_db.h"
#include "../../src/initng_history.h"
#include "../../src/main.h"
#include "../../src/initng_plugin_hook.h"
#include "initng_colorprint_out.h"

#define PE C_RED " %3i%% " C_OFF C_BLUE "%s" C_OFF MOVE_TO_R
#define P C_RED "      " C_OFF C_BLUE "%s" C_OFF MOVE_TO_R


static void out_service_done(char *name)
{
    int time = 0;

    int percent = active_db_percent_started();

    if ((time = time_been_from(name, STARTING)))
        printf(PE "       [" C_GREEN "done" C_OFF
               "]   ( %i seconds to start. )\n", percent, name, time);
    else
        printf(PE "       [" C_GREEN "done" C_OFF "]\n", percent, name);
}


void print_output(active_h * service)
{
    if (!service)
        return;

    /* if, do some output and history adds */
    switch (service->a_status)
    {

        case STOPPED:
            printf(P "    [" C_GREEN "stopped" C_OFF "]\n", service->name);
            return;
        case STARTING:
            printf(PE "   [" C_GREEN "starting" C_OFF "]\n",
                   active_db_percent_started(), service->name);
            return;
        case DONE:
            out_service_done(service->name);
            return;
        case RUNNING:
            if (service->start_process && service->start_process->pid)
                printf(PE "    [" C_GREEN "started" C_OFF
                       "]       (pid : %i).\n", active_db_percent_started(),
                       service->name, service->start_process->pid);
            return;
        case STOP_MARKED:
            printf(P "   [" C_GREEN "stopping" C_OFF "]\n", service->name);
            return;
        case RESPAWNING:
            printf(P "    [" C_GREEN "restart" C_OFF "]\n", service->name);
            return;
        default:
            return;
    }
}


static void print_system_state(h_sys_state state)
{
    switch (state)
    {
        case STATE_STARTING:
            printf("    system is starting up!\n\n");
            break;
        case STATE_STOPING:
            printf("    system is going down!\n\n");
            break;
        case STATE_ASE:
            printf("    Last process exited!\n\n");
            break;
        case STATE_EXIT:
            printf("    Initng exiting!\n\n");
            break;
        case STATE_RESTART:
            printf("    Initng restarting!\n\n");
            break;
        case STATE_UP:
            printf("\n\n"
                   "       runlevel \"%s\" up in, %i seconds.\n\n",
                   s_get_string(S_RUNLEVEL),
                   time_been_from(s_get_string(S_RUNLEVEL), LOADING));
            break;
        default:
            break;
    }
    D_("print_system_state(): new system state: %i\n", state);
}

void print_program_output(active_h * service, int datalen)
{
    /*
       Todo here:
       we should have an internal list of services and "our" current position in them.

       That way when this function is called we can print every full line from plugin_pos.
       This way fsck will look nice, along with an "internal" database of write positions we can cache data so we print every 5 seconds or on int forceflush.
     */



    process_h *x;
    int skipline = FALSE;

    if (datalen == 0)
        return;

    if (service->start_process)
    {
        x = service->start_process;
    }
    else if (service->stop_process)
    {
        x = service->stop_process;
    }
    else
        return;

    int i = x->buffer_pos;

    if (lastservice != service)
    {
        printf(P "       [" C_GREEN "info" C_OFF "]\n", service->name);
        lastservice = service;
    }

    while (x->buffer[i] != '\0')
    {
        if ((x->buffer[i - 1] == '\n') || (x->buffer[i - 1] == '\0'))
            if ((x->buffer[i] == '[') && (x->buffer[i + 2] == ']'))
                skipline = TRUE;

        if (!skipline)
            putchar(x->buffer[i]);
        if (x->buffer[i] == '\n')
            skipline = FALSE;
        i++;
    }
    /*    if (x->buffer[i - 1] != '\n') {
       putchar('\n');
       //        printf("End data : %d, %d, %d, %d, '%c' '%c' '%c' '%c'\n",x->buffer[i],x->buffer[i-1],x->buffer[i-2],x->buffer[i-3],x->buffer[i],x->buffer[i-1],x->buffer[i-2],x->buffer[i-3]);
       } */
}

int module_init(void)
{
    printf("\n" C_BLUE "     Next Generaton Init version ( ");
    printf(INITNG_VERSION);
    printf(" )" C_OFF);
    printf("\n" C_GREEN "     http://jw.dyndns.org/initng/\n" C_OFF);
    printf("     Author: Jimmy Wennlund <jimmy.wennlund@gmail.com>\n\n");

    D_("module_init();\n");
    lastservice = NULL;
    initng_add_hook(WATCHERS, &print_output);
    initng_add_hook(SWATCHERS, &print_system_state);
    initng_add_hook(PIPEWATCHERS, &print_program_output);
    return (TRUE);
}

void module_unload(void)
{
    D_("color_out: moule_unload();\n");
    initng_del_hook(WATCHERS, &print_output);
    initng_del_hook(SWATCHERS, &print_system_state);
    initng_del_hook(PIPEWATCHERS, &print_program_output);
    printf("  Goodbye\n");
}
