
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#ifdef BUSYBOX
#include "../../busybox-1.00/include/busybox.h"
#include "../../busybox-1.00/include/libbb.h"
#endif
#include "../../src/initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <stdio.h>
#include <string.h>                         /* memset() strlen() */
#include <errno.h>

#include "../../src/initng_execute.h"       /* new_environ() */
#include "../../src/initng_global.h"
#include "../../src/initng_toolbox.h"

#include "dash_executor.h"

/* dash block : */
#include "shell.h"
#include "main.h"
#include "options.h"
#include "output.h"
#include "parser.h"
#include "expand.h"
#include "eval.h"
#include "jobs.h"
#include "input.h"
#include "trap.h"
#include "var.h"
#include "show.h"
#include "memalloc.h"
#include "error.h"
#include "init.h"
#include "mystring.h"
#include "exec.h"
#include "cd.h"

static void do_kill(int sig_num);

int rootpid;
int rootshell;
#ifdef __GLIBC__
int *dash_errno;
#endif

/*extern char **environ; */
int status = 0;                 /* exit status */
int running = TRUE;             /* Process is running */
pid_t pid_fork = 0;

static void do_kill(int sig_num)
{
    D_("do_kill(): #####################    Got signal: %i   Forwarding to %i!\n", sig_num, pid_fork);
    signal(sig_num, do_kill);
    kill(pid_fork, sig_num);
}

static void dash_exited()
{
    if (wait(&status) == -1)
        F_("Got error from wait %s", strerror(errno));
    running = FALSE;
}

void execute_dash(char *dash_code)
{
    struct jmploc jmploc;
    volatile int state;

    if (setjmp(jmploc.loc))
    {
        int status;
        int e;

        reset();

        e = exception;
        switch (exception)
        {
            case EXEXEC:
                status = exerrno;
                break;
            case EXERROR:
                status = 2;
                break;
            default:
                status = exitstatus;
                break;
        }
        exitstatus = status;
        if (e == EXEXIT || state == 0 || iflag == 0 || !rootshell)
            exitshell();
    }
    handler = &jmploc;
    evalstring(dash_code);
    exitshell();
}

int dash_initialize()
{
#ifdef __GLIBC__
    dash_errno = __errno_location();
#endif
    struct jmploc jmploc;

    if (setjmp(jmploc.loc))
    {
        int status;
        int e;

        reset();
    }
    handler = &jmploc;

    rootpid = getpid();
    rootshell = 1;
}

void setupvars(char **environ)
{
    char **envp;

#define VEXPORT		0x01                /* variable is exported */
#define VTEXTFIXED	0x08                /* text is statically allocated */
    for (envp = environ; *envp; envp++)
    {
        if (strchr(*envp, '='))
        {
            setvareq(*envp, VEXPORT | VTEXTFIXED);
        }
    }
}

/* Fake functions so we can link without errors: */

void cmdloop(int top)
{
}
STATIC void read_profile(const char *name)
{
}
void readcmdfile(char *name)
{
}
STATIC char *find_dot_file(char *basename)
{
}
int dotcmd(int argc, char **argv)
{
}

int exitcmd(int argc, char **argv)
{
    if (stoppedjobs())
        return 0;
    if (argc > 1)
        exitstatus = number(argv[1]);
    exraise(EXEXIT);
    /* NOTREACHED */
}

void dash_this(char *dash_code, active_h * s)
{
    int inpipe_fork[2];
    char buffer[512];
    int fd_flags;
    int timerun = 0;
    int len = 0;
    int header_printed = 0;

    D_("dash_this(%s);\n", s->name);
    pipe(inpipe_fork);
    
    /* send these signals to child if we got them */
    signal(SIGTERM, do_kill);
    signal(SIGCHLD, dash_exited);
    /* signal(SIGKILL, do_kill); kill is not set */


    switch (pid_fork = fork())
    {

        case -1:
            _exit(34);
        case 0:                            /* child */
            D_("dash_this(): This is child: %i\n", pid_fork);
            /* close original fd's, copy fd's to our pipes and close unused sides on pipes. */
            close(1);
            dup2(inpipe_fork[1], 1);
            close(2);
            dup2(inpipe_fork[1], 2);
            close(inpipe_fork[0]);
            setupvars(new_environ(s));
            commandname = s->name;
            signal(SIGCHLD, SIG_DFL);       /* avoids infinite loop bug. */
            execute_dash(dash_code);
            /* NEVER REACHED */
            F_("dash_this(): child died!\n EEEEEEEEEEEEEEEEEEEEEEERRRRRRRRRRRROOOOOOOOOORR!\n");
            /* exit 1, why???, to emit signal? */
            initng_free();
            _exit(1);
        default:                           /* non fork comes here */
            /* if we need to wait for it */
            D_("dash_this(): This is parent to: %i\n", pid_fork);

            /* Read data from pipe : */
            close(inpipe_fork[1]);
            fd_flags = fcntl(inpipe_fork[0], F_GETFL, 0);
            fcntl(inpipe_fork[0], F_SETFL, fd_flags | O_NONBLOCK);

            /* Check for new output every RUN_OUTPUTINFO seconds */
            do
            {
                if (running)
                {
                    timerun++;
                    usleep(1000000);
                }
                header_printed = 0;
                while ((len = read(inpipe_fork[0], &buffer, 511)) > 0)
                {
                    buffer[len] = '\0';
                    /* Data is in buffer */
                    if (header_printed == 0)
                    {
                        header_printed = 1;
                        printf("\n-- \"%s\" (%i) :\n    ", s->name, pid_fork);
                    }
                    len = 0;
                    /* Print output buffer */
                    while (buffer[len] != '\0' && len < 512)
                    {
                        putchar(buffer[len]);
                        if (buffer[len] == '\n')
                            printf("    ");
                        len++;
                    }
                }
                if (timerun == RUN_TIMEOUT_WARN)
                {
                    printf
                        ("\n-- WARNING %s has been running for %d (%i) seconds, last output given :\n",
                         s->name, pid_fork, RUN_TIMEOUT_WARN);
                    printf(buffer);
                }
            }
            while (running);
            close(inpipe_fork[1]);

            D_("dash_this(): TODO, tell nginit the pid of this process is %i!\n", pid_fork);

#ifdef DEBUG
            if (pid_fork > 0)
            {
                D_("dash_executor(): PID %i KILLED WITH:\n"
                   "WEXITSTATUS %i\n" "WIFEXITED %i\n" "WIFSIGNALED %i\n"
                   "WTERMSIG %i\n" "WCOREDUMP %i\n" "WIFSTOPPED %i\n"
                   "WSTOPSIG %i\n", pid_fork,
                   WEXITSTATUS(status), WIFEXITED(status),
                   WIFSIGNALED(status), WTERMSIG(status),
                   WCOREDUMP(status), WIFSTOPPED(status), WSTOPSIG(status));
            }
#endif
            initng_free();
            _exit(WEXITSTATUS(status));
    }                                       /* End case */
}                                           /* end fork_and_exec() */
