
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <stdio.h>
#include <string.h>                         /* strdup() */
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>

#include "../../src/active_db.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_execute.h"
#include "../../src/string_tools.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_signal.h"

#include "dash_executor.h"

#include "initng_dash_launcher.h"

static pid_t
script_initng_execute_fork_and_exec(process_h * process_to_exec, active_h * s,
                                    char *script);

int initng_dash_start(active_h * service)
{
    char *e = NULL;

    assert(service);
    assert(service->from_service);

    /* WE ARE EXECUTING A START SCRIPT */
    if ((e = active_db_get_string(START_SCRIPT, service)))
    {
        if (script_initng_execute_fork_and_exec
            (service->start_process, service, e))
            return (TRUE);

        return (FAIL);
    }
    return (FALSE);
}

int initng_dash_stop(active_h * service)
{
    char *e = NULL;

    assert(service);
    assert(service->from_service);

    /* WE ARE EXECUTING A START SCRIPT */
    if ((e = active_db_get_string(STOP_SCRIPT, service)))
    {
        if (script_initng_execute_fork_and_exec
            (service->stop_process, service, e))
            return (TRUE);

        return (FAIL);
    }
    return (FALSE);

}


static pid_t
script_initng_execute_fork_and_exec(process_h * process_to_exec, active_h * s,
                                    char *script)
{

    /* called from inside the service directory, return the PID or 0 on error */

    /* This is the real service kicker */
    pid_t pid_fork;             /* pid got from fork() */
    int try_count = 0;          /* Count tryings */

    assert(process_to_exec);
    assert(script);

    /* Try to fork 3 times */
    while ((pid_fork = fork()) == -1)
    {
        if (try_count > 3)
            return 0;                       /* Already tried 3 times = no more try */
        sleep(++try_count * 2);             /* Sleep a while before trying again */
    }

    /* if this is child */
    if (pid_fork == 0)
    {
        char *tmp = NULL;
        int t = 0;

	disable_signals();
	
        /* we dont need the ngc fifos in fork */
        /*remote_control_close(); */

        if (active_db_is(NICE, s))
        {
            nice(active_db_get_int(NICE, s));
        }


	if(g.i_am_init) {
            ioctl(0, TIOCNOTTY, 0);
            setsid();                       /* Run a program in a new session ??? */
	}
	
	if (active_db_is(STDOUT, s) || active_db_is(STDERR, s)) {
	    open_requested_out(s);
	}
        else if (g.i_am_init)
        {                                   /* If this is init */
            open_dev_out();
        }
	
	if(g.i_am_init)
            tcsetpgrp(0, getpgrp());        /* run this in foreground on fd 0 */

        if ((tmp = active_db_get_string(CHDIR, s)))
        {
            D_("change dir to: %s\n", tmp);
            chdir(tmp);
        }


        if ((tmp = active_db_get_string(CHROOT, s)))
        {
            D_("change root to: %s\n", tmp);
            chdir(tmp);
            chroot(tmp);
        }

        if ((tmp = active_db_get_string(SUID, s)))
        {
            struct passwd *pw;

            pw = getpwnam(tmp);
            if (!pw)
                F_("%s: USER %s not found!", s->name, tmp);
            D_("setuid to: %i.\n", pw->pw_uid);
            setuid(pw->pw_uid);
        }

        /* delay execution a bit */
        if ((t = active_db_get_int(UDELAY, s)) > 0)
        {
            D_("Sleeping for %i miliseconds.\n", t);
            usleep(t);
        }
        else if ((t = active_db_get_int(DELAY, s)) > 0)
        {
            D_("Sleeping for %i seconds.\n", t);
            sleep(t);
        }
        else
        {
            /* do a minimum sleep, to let the mother process 
               to register child, and notice death */
            usleep(ALL_USLEEP);
        }

        /* execute code */
        dash_this(script, s);

        /* dash This should NOT return! */
        _exit(33);
    }

    /* save pid of fork */
    D_("FROM_FORK Forkstarted pid %i.\n", pid_fork);
    process_to_exec->pid = pid_fork;
    return pid_fork;

}

int module_init(void)
{
    D_("initng_simple_plugin: module_init();\n");

    dash_initialize();

    initng_add_hook(STARTL, &initng_dash_start);
    initng_add_hook(STOPL, &initng_dash_stop);

    return (TRUE);
}

void module_unload(void)
{
    D_("initng_simple_plugin: module_unload();\n");

    initng_del_hook(STARTL, &initng_dash_start);
    initng_del_hook(STOPL, &initng_dash_stop);

}
