
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>

#define XML_EXT ".xml"
#include "../../src/string_tools.h"
#include "../../src/service_db.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"

#include "initng_ezxml_parser.h"
#include "ezxml.h"

static int set_xml_parser(e_sdt type, char *value, service_h * from_service);
static int unset_xml_parser(e_sdt type, char *value,
                            service_h * from_service);
static int string_xml_parser(e_sdt type, char *value,
                             service_h * from_service);
static int strings_xml_parser(e_sdt type, char *value,
                              service_h * from_service);
static int int_xml_parser(e_sdt type, char *value, service_h * from_service);

static int set_xml_parser(e_sdt type, char *value, service_h * from_service)
{
    assert(from_service);
    if (!value)
        return (FALSE);

    D_("set_xml_parser(%i,%s,%s);\n", type, value, from_service->name);
    service_db_set(type, from_service);
    return (TRUE);
}

static int unset_xml_parser(e_sdt type, char *value, service_h * from_service)
{
    assert(from_service);
    if (!value)
        return (FALSE);
    D_("unset_xml_parser(%i,%s,%s);\n", type, value, from_service->name);

    service_db_remove(type, from_service);
    return (TRUE);
}


/* funtion copy full string into db */
static int string_xml_parser(e_sdt type, char *value,
                             service_h * from_service)
{
    assert(from_service);
    if (!value)
        return (FALSE);
    D_("string_parser(%i,%s,%s);\n", type, value, from_service->name);
    service_db_set_string(type, from_service, strdup(value));
    return (FALSE);
}


static int int_xml_parser(e_sdt type, char *value, service_h * from_service)
{
    assert(from_service);
    if (!value)
        return (FALSE);
    D_("string_parser(%i,%s,%s);\n", type, value, from_service->name);
    service_db_set_int(type, from_service, atoi(value));
    return (FALSE);
}

/* funtion splits up all string to single entrys */
static int strings_xml_parser(e_sdt type, char *value,
                              service_h * from_service)
{
    char *to = NULL;

    assert(from_service);
    if (!value)
        return (FALSE);
    D_("strings_xml_parser(%i,%s,%s);\n", type, value, from_service->name);

    while ((to = st_dup_next_word(&value)))
    {
        service_db_set_another_string(type, from_service, to);
        st_jump_spaces(&value);
    }
    return (TRUE);
}

static int parse_xml_opt(char *name, char *value, service_h * srv)
{
    int i;

    for (i = 0; g.option_table[i].opt_name != NULL; i++)
    {
        if (strcasecmp(g.option_table[i].opt_name, name) == 0)
        {
            switch (g.option_table[i].opt_type)
            {
                case STRING:
                    return (string_xml_parser
                            (g.option_table[i].opt_id, value, srv));
                case STRINGS:
                    return (strings_xml_parser
                            (g.option_table[i].opt_id, value, srv));
                case SET:
                    return (set_xml_parser
                            (g.option_table[i].opt_id, value, srv));
                case UNSET:
                    return (unset_xml_parser
                            (g.option_table[i].opt_id, value, srv));
                case SCRIPT:
                    return (string_xml_parser
                            (g.option_table[i].opt_id, value, srv));
                case INT:
                    return (int_xml_parser
                            (g.option_table[i].opt_id, value, srv));
                case U_D_T:
                    break;
            }

            D_("parse_xml_opt(%s,%s): at option no %i.\n", name, value, i);
        }

    }

    F_("parse_xml_opt(%s): error line: \"%s\"!\n", name, value);
    F_("sleeping 2 seconds......\n");
    sleep(2);
    return (FALSE);
}

static service_h *initng_parse_service_ezxml(ezxml_t node, char *watch_for,
                                             service_h * father, e_type type)
{
    char *name = NULL;
    service_h *copy = NULL;
    service_h *new_service = NULL;
    service_h *service_to_return = NULL;

    assert(watch_for);
    if (!node)
        return (NULL);

    D_("initng_parse_service_ezxml();\n");

    if (!node->attr[0] ||
        !strcasecmp(node->attr[0], "name") == 0 ||
        !(node->attr[1]) || !(name = strdup(node->attr[1])))
    {
        F_("initng_parse_service_ezxml(): service must have a name!");
        return (NULL);
    }

    D_("initng_parse_service_ezxml(): service name: \"%s\"\n", name);

    /* create a service entry */
    if (node->attr[2] &&
        strcasecmp(node->attr[2], "father") == 0 && node->attr[3])
    {
        copy = service_db_find_by_name((char *) node->attr[3]);
        if (copy)
            service_db_copy(&new_service, name, copy);
    }
    else
    {
        service_db_new(&new_service, name, type);
    }

    if (!new_service)
    {
        F_("initng_parse_service_ezxml(): Unable to allocate service.\n");
        return (NULL);
    }

    /* set type if not set. */
    if (father)
        new_service->type = type;

    node = node->child;

    while (node)
    {
        service_h *tmp = NULL;

        /*
           if(strcmp((char *)node->name,"text")==0) {
           node=node->next;    
           continue;
           }
         */
        D_("node-name:%s\n", node->name);
        if (strcasecmp("service", (char *) node->name) == 0)
            if ((tmp =
                 initng_parse_service_ezxml(node->child, watch_for, NULL,
                                            TYPE_SERVICE)))
                service_to_return = tmp;

        if (strcasecmp("class", (char *) node->name) == 0)
            if ((tmp =
                 initng_parse_service_ezxml(node->child, watch_for, NULL,
                                            TYPE_CLASS)))
                service_to_return = tmp;

        /* parse line - search for keywords in g.option_table */
        parse_xml_opt((char *) node->name, node->txt, new_service);

        /* go to next node */
        node = node->ordered;
    }

    /* add to service db */
    if (!service_db_add(new_service))
    {
        F_("can't add service to service db!\n");
        if (new_service->name)
            free(new_service->name);
        if (new_service)
            free(new_service);
        return (NULL);
    }

    /* check if this is the service we ar looking for */
    if (st_service_cmp(watch_for, new_service->name))
        return (new_service);

    /* if we got a service from a under service */
    if (service_to_return)
        return (service_to_return);
    /* return null if not */
    return (NULL);

}

/* parse a file for services */
static service_h *initng_parse_ezxml_file(char *filename, char *watch_for)
{
    service_h *service_to_return = NULL;
    service_h *tmp = NULL;

    int file_fd;
    ezxml_t xml_file;

    assert(filename);
    assert(watch_for);
    D_("parse_xml_file(%s);\n", filename);


    if ((file_fd = open(filename, O_RDONLY, 0)) == -1)
    {
        D_("parse_ezxml_file(%s): Can't open xml file for parsing.\n",
           filename);
        return (NULL);
    }

    /* open xml file */
    if (!(xml_file = ezxml_parse_fd(file_fd)))
    {
        D_("parse_ezxml_file(%s): Can't parse file descriptor.\n", filename);
        close(file_fd);
        return (NULL);
    }

    /* big loop */
    while (xml_file)
    {
        D_("headnode: %s\n", xml_file->name);
        if (strcasecmp("service", xml_file->name) == 0)
        {
            if ((tmp =
                 initng_parse_service_ezxml(xml_file, watch_for, NULL,
                                            TYPE_SERVICE)))
                service_to_return = tmp;

        }
        else if (strcasecmp("class", xml_file->name) == 0)
        {
            if ((tmp =
                 initng_parse_service_ezxml(xml_file, watch_for, NULL,
                                            TYPE_CLASS)))
                service_to_return = tmp;
        }
        xml_file = xml_file->ordered;

    }


    /*exit_initng(1); */
    close(file_fd);
    if (service_to_return)
        return (service_to_return);
    return (NULL);
}

/* Load a service from a process_name or process_path */
service_h *initng_ezxml_load_process(char *process_name)
{
    char filetoparse[200];
    char *tmp = NULL;
    service_h *got_serv = NULL;

    assert(process_name);
    /* example  INITNG_ROOT=/etc/initng process_name="deamon/sshd XML_EXT=".xml" */
    /* TRY NO 1, try load   "/etc/initng/daemon/sshd.xml" */
    /* TRY NO 2, try load   "/etc/initng/daemon/sshd/sshd.xml" */
    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.xml" */
    /* TRY NO 4, try load   "/etc/initng/daemon/default.xml" */
    /* TRY NO 5, try load   "/etc/initng/daemon/daemon.xml */
    /* TRY NO 6, try load   "/etc/initng/sshd.xml" */
    /* TRY NO 7, try load   "/etc/initng/daemon.xml */

    /* TRY NO 1, try load   "/etc/initng/daemon/sshd/sshd.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 2, try load   "/etc/initng/daemon/sshd.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 3, try load   "/etc/initng/daemon/sshd/default.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, process_name);
    strcat(filetoparse, "/default" XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 4, try "/etc/initng/daemon/default.xml" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/default" XML_EXT);
    free(tmp);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 5, try "/etc/initng/daemon/daemon.xml" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, XML_EXT);
    free(tmp);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 6, try "/etc/initng/sshd.xml" */
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, st_strip_path(process_name));
    strcat(filetoparse, XML_EXT);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    /* TRY NO 7, try "/etc/initng/daemon.xml" */
    tmp = st_get_path(process_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, tmp);
    strcat(filetoparse, XML_EXT);
    free(tmp);
    if ((got_serv = initng_parse_ezxml_file(filetoparse, process_name)))
        return (got_serv);

    D_("initng_ezxml_load_process(%s): got \"NULL\"\n", process_name);
    return (NULL);
}


int module_init(void)
{
    D_("moule_init();\n");
    initng_add_hook(PARSERS, &initng_ezxml_load_process);
    return (TRUE);
}

void module_unload(void)
{

    D_("moule_unload();\n");
    initng_del_hook(PARSERS, &initng_ezxml_load_process);

}
