
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#define _GNU_SOURCE
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <getopt.h>

/* #include "ngc.h" */
#include "../../src/initng.h"
#include "../../src/active_status.h"

#define TRUE 1
#define FALSE 0

static struct option options[] = {
    {"get", required_argument, NULL, 'g'},
    {"up", required_argument, NULL, 'u'},
    {"down", required_argument, NULL, 'd'},
    {"father", required_argument, NULL, 'f'},
    {"reload", required_argument, NULL, 'r'},
    {"j", required_argument, NULL, 'j'},
    {"zap", required_argument, NULL, 'z'},
    {"log", no_argument, NULL, 'l'},
    {"status", no_argument, NULL, 's'},
    {"quit", no_argument, NULL, 'q'},
    {"restart", no_argument, NULL, 'x'},
    {"print", optional_argument, NULL, 'p'},
    {"print_service", required_argument, NULL, 'k'},
    {"startfrom", required_argument, NULL, 'j'},
    {"n", no_argument, NULL, 'n'},
    {"help", no_argument, NULL, 'h'},

};

/* Global in and out fifos */
static int send_pipe, recv_pipe;

/* functions */
void print_usage(void);
int open_pipes(void);
int close_pipes(void);
char *command_fetch_string(char *command);
unsigned int command_fetch_int(char *command);
unsigned int print_status(char *seervicename, e_a_status status,
                          struct tm *ts);
int main(int argc, char *argv[]);

unsigned int ngc_get_pid(char *service);
unsigned int ngc_take_up(char *service);
unsigned int ngc_take_down(char *service);
char *ngc_father_to(char *service);
unsigned int ngc_reload(char *service);
unsigned int ngc_restart_from(char *service);
unsigned int ngc_zap(char *service);
unsigned int ngc_history(void);
unsigned int ngc_status(void);
unsigned int ngc_restart(void);
unsigned int ngc_quit(void);
unsigned int ngc_print(char *service);
unsigned int ngc_percent(void);
unsigned int ngc_close_io(void);
unsigned int ngc_togle_verbose(void);

/* print usage on screen */
void print_usage(void)
{
    printf("\nusage: ngc -[gudfrj] service\n"
           "       ngc -[lsxqpn]\n\n"
           " -g\tget; output just the PID.\n"
           " -u\tup; start service.\n"
           " -d\tdown; stop service.\n"
           " -f\tfather; find out father to process.\n"
           " -r\treload; service data from disk.\n"
           " -j\tstartfrom, stops all services, and start just this one.\n");
    printf(" -z\tzap, mark it as stop, and removed from active services.\n"
           " -l\thistory; print last history.\n"
           " -s\tstatus; List all active processes.\n"
           " -q\tquit; kill initng (hang).\n"
           " -x\trestart; initng (warm reboot).\n"
           " -p\tprint; Dump service cache on console.\n"
           " -v\tverbose; Toggle verbosity.\n"
           " -n\tprint percent; For how many percent is your system up?.\n\n");
}

/* open pipes to comunicate with */
int open_pipes(void)
{
    int i = 20;                 /* counter */

    /* open pipes */
#ifdef NGDC
    send_pipe = open(CONTROL_PATH "/in-test", O_WRONLY);
    recv_pipe = open(CONTROL_PATH "/out-test", O_RDONLY);
#else
    send_pipe = open(CONTROL_PATH "/in", O_WRONLY);
    recv_pipe = open(CONTROL_PATH "/out", O_RDONLY);
#endif
    /* check pipes */
    if (send_pipe < 0 || recv_pipe < 0)
    {
        printf("ngc: could not open " CONTROL_PATH "/in or " CONTROL_PATH
               "/out!\n");
        return 1;
    }

    /* TRY SET LOCK, so there is not another program talking to initng */
    while (lockf(send_pipe, F_LOCK, 1))
    {
        printf("ngc: could not acquire lock!  trying .... %i...\n", i);
        i--;
        if (i == 0)
        {
            printf("ngc: error giving up!\n");
            return (1);
        }
        sleep(1);
    }
    return 0;
}

/* close pipes */
int close_pipes(void)
{
    return (close(send_pipe) + close(recv_pipe));
}

/* send a command return the fetched string */
char *command_fetch_string(char *command)
{
    char *got_from_read;

    if (open_pipes() == 1)
        return (NULL);
    got_from_read = calloc(10001, sizeof(char));
    write(send_pipe, command, strlen(command));
    /* no idea to stress initng */
    usleep(100000);
    got_from_read[read(recv_pipe, got_from_read, 10000)] = '\0';
    if (close_pipes())
        return (NULL);
    return (got_from_read);
}

/* send command and return the fetched return number */
unsigned int command_fetch_int(char *command)
{
    char got_from_read[20];

    if (open_pipes() == 1)
        return (-1);
    write(send_pipe, command, strlen(command));
    /* no idea to stress initng */
    usleep(50000);
    got_from_read[read(recv_pipe, &got_from_read, 19)] = '\0';
    if (close_pipes())
        return (-1);
    return (atoi(got_from_read));
}

/* THIS IS MAIN */
int main(int argc, char *argv[])
{
    int c = 0;
    int success = 0;

    /* print banner */
    printf("Next Generation init Control.\n");
    printf("written by Jimmy Wennlund <jimmy.wennlund@gmail.com>\n\n");

    while (1)
    {
        int option_index = 0;

        c = getopt_long(argc, argv, "g:u:d:f:r:j:z:k:p::lsxqhniv", options,
                        &option_index);
        if (c == -1)
            break;
        success++;                          /* will not display help on return */
        switch (c)
        {
            case 'l':                      /* get history */
                ngc_history();
                break;
            case 's':                      /* get running processes */
                ngc_status();
                break;
            case 'x':                      /* restart initng */
                ngc_restart();
                break;
            case 'q':                      /* restart initng */
                ngc_quit();
                break;
            case 'n':                      /* print Percent done */
                ngc_percent();
                break;
            case 'i':                      /* print Percent done */
                ngc_close_io();
                break;
            case 'v':                      /* toggle verbosity */
                ngc_togle_verbose();
                break;
            case 'p':                      /* initng print */
                ngc_print(optarg);
                break;
            case 'k':                      /* initng print */
                ngc_print(optarg);
                break;
            case 'g':                      /* get pid of */
                ngc_get_pid(optarg);
                break;
            case 'u':                      /* start service */
                ngc_take_up(optarg);
                break;
            case 'd':                      /* take service down */
                ngc_take_down(optarg);
                break;
            case 'f':                      /* Find out father to */
                ngc_father_to(optarg);
                break;
            case 'r':                      /* reload service data from disk */
                ngc_reload(optarg);
                break;
            case 'j':                      /* restart from */
                ngc_restart_from(optarg);
                break;
            case 'z':                      /* zap */
                ngc_zap(optarg);
                break;
            case 'h':                      /*  help */
            default:
                print_usage();
                break;
        }                                   /* end switch */
    }

    /* return happily */
    if (!success)
    {
        print_usage();
        return (1);
    }
    return 0;
}

/* get pid of */
unsigned int ngc_get_pid(char *service)
{
    char write_this[101];
    unsigned int pid = 0;

    sprintf(write_this, "g%s", service);
    pid = command_fetch_int(write_this);
    if (pid > 0)
    {
        printf("pid of %s is: %i\n", service, pid);
        return (pid);
    }
    else
    {
        printf("ngc: Error getting pid of %s.\n", service);
        return (0);
    }
}

/* start service */
unsigned int ngc_take_up(char *service)
{
    char write_this[101];
    unsigned int s;

    sprintf(write_this, "u%s", service);
    if ((s = command_fetch_int(write_this)) == 1)
    {
        printf("Process %s started. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to start process %s. %i :-(\n", service, s);
        return (FALSE);
    }
}

/* take service down */
unsigned int ngc_take_down(char *service)
{
    char write_this[101];

    sprintf(write_this, "d%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("Process %s stopped. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to stop process %s. :-(\n", service);
        return (FALSE);
    }
}

/* restart from */
unsigned int ngc_restart_from(char *service)
{
    char write_this[101];

    sprintf(write_this, "j%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("restart from %s ok. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to restart from process %s. :-(\n", service);
        return (FALSE);
    }
}

/* zap action */
unsigned int ngc_zap(char *service)
{
    char write_this[101];

    sprintf(write_this, "z%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("zap %s ok. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to zap process %s. :-(\n", service);
        return (FALSE);
    }
}

unsigned int print_status(char *servicename, e_a_status status, struct tm *ts)
{

    /* Print right message by messagetype */
    switch (status)
    {

        case NONE:
            printf("          %s : No status, maybe a new service.\n",
                   servicename);
            break;
        case LOADING:
            printf(" %.2i:%.2i:%.2i %s : loading.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case STARTING:
            printf(" %.2i:%.2i:%.2i %s : starting.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case WAITING_FOR_START_DEP:
            printf(" %.2i:%.2i:%.2i %s : waiting for dependency to start.\n",
                   ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case START_DEP_MET:
            printf(" %.2i:%.2i:%.2i %s : start dependency met, starting.\n",
                   ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case FAIL_STARTING:
            printf(" %.2i:%.2i:%.2i %s : FAILD starting.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case RUNNING:
            printf(" %.2i:%.2i:%.2i %s : running.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case DONE:
            printf(" %.2i:%.2i:%.2i %s : done.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case RESPAWNING:
            printf(" %.2i:%.2i:%.2i %s : respawning.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOPPING:
            printf(" %.2i:%.2i:%.2i %s : stopping.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case WAITING_FOR_STOP_DEP:
            printf
                (" %.2i:%.2i:%.2i %s : is waiting for depenencys to stop.\n",
                 ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOP_DEP_MET:
            printf
                (" %.2i:%.2i:%.2i %s : all stop dependencys are met, now stopping\n",
                 ts->tm_hour, ts->tm_min, ts->tm_sec, servicename);
            break;
        case FAIL_STOPPING:
            printf(" %.2i:%.2i:%.2i %s : stopping.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        case STOPPED:
            printf(" %.2i:%.2i:%.2i %s : stopped.\n", ts->tm_hour, ts->tm_min,
                   ts->tm_sec, servicename);
            break;
        case STOP_MARKED:
            printf(" %.2i:%.2i:%.2i %s : marked stop.\n", ts->tm_hour,
                   ts->tm_min, ts->tm_sec, servicename);
            break;
        default:
            printf(" %s : error!\n", servicename);
            return (FALSE);
            break;
    }

    return (TRUE);

}

/* Find out father to */
char *ngc_father_to(char *service)
{
    char write_this[101];
    char *got_this;

    sprintf(write_this, "f%s", service);
    got_this = command_fetch_string(write_this);
    if (got_this)
    {
        printf("Father to %s is %s.\n", service, got_this);
        return (got_this);
    }
    else
    {
        printf("ngc: Error getting father or father unknown.\n");
        return (NULL);
    }
}

/* reload service data from disk */
unsigned int ngc_reload(char *service)
{
    char write_this[101];

    sprintf(write_this, "r%s", service);
    if (command_fetch_int(write_this) == 1)
    {
        printf("Service %s data reloaded. :-)\n", service);
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to reload %s. :-(\n", service);
        return (FALSE);
    }
}

/* get history */
unsigned int ngc_history(void)
{
    char *got_this;
    unsigned int i = 0;
    unsigned int j = 0;
    int type;
    long int t;
    struct tm *ts;
    char servicename[41];

    /* Fetch input from initng */
    if (!(got_this = command_fetch_string("h")))
        return (FALSE);

    /* print head */
    printf(" hh:mm:ss service (fetching a %.5i byte table)    : status\n",
           (int) strlen(got_this));
    printf(" ----------------------------------------------------------\n");

    /* as long as there is something to read left */
    while (i <= (strlen(got_this) - 2))
    {

        /* Parse the line into type and servicename */
        if (!sscanf(got_this + i, "%i|%li%s\n", &type, &t, servicename))
            return (FALSE);
        {
            time_t test = (time_t) t;

            ts = localtime(&test);
        }
        /*timen=(time_t)t; */

        /* Step i forward to next entry */
        i = i + strcspn(got_this + i, "\n\0") + 1;

        /* Fill out 40 blanks for nice output */
        for (j = strlen(servicename); j <= 40; j++)
            servicename[j] = ' ';
        servicename[40] = '\0';

        print_status(servicename, type, ts);
    }
    return (TRUE);
}

/* get running processes */
unsigned int ngc_status(void)
{
    char *got_this;
    unsigned int i = 0;
    unsigned int j = 0;
    long int ti;
    struct tm *ts;
    int type;
    char servicename[41];

    /* Fetch input from initng */
    if (!(got_this = command_fetch_string("s")))
        return (FALSE);

    /* print head */
    printf(" hh:mm:ss service (fetching a %.5i byte table)    : status\n",
           (int) strlen(got_this));
    printf(" ----------------------------------------------------------\n");

    /* as long as there is something to read left */
    while (i <= (strlen(got_this) - 2))
    {

        /* Parse the line into type and servicename */
        if (!sscanf(got_this + i, "%i|%li|%s\n", &type, &ti, servicename))
            break;

        /* Step i forward to next entry */
        i = i + strcspn(got_this + i, "\n\0") + 1;
        {
            time_t test = (time_t) ti;

            ts = localtime(&test);
        }
        /* Fill out 40 blanks for nice output */
        for (j = strlen(servicename); j <= 40; j++)
            servicename[j] = ' ';
        servicename[40] = '\0';

        print_status(servicename, type, ts);
    }
    return (TRUE);
}

/* restart initng */
unsigned int ngc_restart(void)
{
    if (command_fetch_int("x") == 1)
    {
        printf("Initng reloading itself now...\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to reload initng. :-(\n");
        return (FALSE);
    }
}

/* restart initng */
unsigned int ngc_togle_verbose(void)
{
    int a = command_fetch_int("v");

    if (a == 2)
    {
        printf("Verbose is on...\n");
        return (TRUE);
    }
    else if (a == 1)
    {
        printf("Verbose is off...\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to toggle verbose mode. :-(\n");
        return (FALSE);
    }
}

/* quit initng (system will hang) */
unsigned int ngc_quit(void)
{
    if (command_fetch_int("q") == 1)
    {
        printf("Initng is quiting...\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Failed to quit initng. :-(\n");
        return (FALSE);
    }
}

/* Fetch Percent */
unsigned int ngc_percent(void)
{
    int percent;

    percent = command_fetch_int("n");
    printf("Percent done %i.\n", percent);
    exit(percent);
}

/* Fetch Percent */
unsigned int ngc_close_io(void)
{
    exit(command_fetch_int("i"));
}

/* Print services on daemon out */
unsigned int ngc_print(char *service)
{
    char write_this[101];
    int res;

    if (service)
    {
        printf("\"p%s\"", service);
        sprintf(write_this, "p%s", service);
        res = command_fetch_int(write_this);
    }
    else
        res = command_fetch_int("p");

    if (res == 1)
    {
        printf("Good...\n");
        return (TRUE);
    }
    else
    {
        printf("ngc: Bad. :-(\n");
        return (FALSE);
    }
}
