
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* purpose: argv[1] is the full path to a PID file,
 *          argv+2 is the daemon to run.
 * the daemon is expected to fork in the background and write its PID in
 * the pid file.
 */

#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "../../src/initng.h"

#define TRUE 1
#define FALSE 0

/* Global in and out fifos */
static int send_pipe, recv_pipe;
extern char **environ;

int open_pipes(void);
int close_pipes(void);
int command_fetch_int(char *command);

/* open pipes to comunicate with */
int open_pipes(void)
{
    int i = 20;                 /* counter */

    /* open pipes */
#ifdef NGDC
    send_pipe = open(CONTROL_PATH "/in-test", O_WRONLY);
    recv_pipe = open(CONTROL_PATH "/out-test", O_RDONLY);
#else
    send_pipe = open(CONTROL_PATH "/in", O_WRONLY);
    recv_pipe = open(CONTROL_PATH "/out", O_RDONLY);
#endif
    /* check pipes */
    if (send_pipe < 0 || recv_pipe < 0)
    {
        printf("ngc: could not open " CONTROL_PATH "/in or " CONTROL_PATH
               "/out!\n");
        return 1;
    }

    /* TRY SET LOCK, so there is not another program talking to initng */
    while (lockf(send_pipe, F_LOCK, 1))
    {
        printf("ngc: could not acquire lock!  trying .... %i...\n", i);
        i--;
        if (i == 0)
        {
            printf("ngc: error giving up!\n");
            return (1);
        }
        sleep(1);
    }
    return 0;
}

/* close pipes */
int close_pipes(void)
{
    return (close(send_pipe) + close(recv_pipe));
}

/* send command and return the fetched return number */
int command_fetch_int(char *command)
{
    char got_from_read[20];

    if (open_pipes() == 1)
        return (-1);
    write(send_pipe, command, strlen(command));
    /* no idea to stress initng */
    usleep(50000);
    got_from_read[read(recv_pipe, &got_from_read, 19)] = '\0';
    if (close_pipes())
        return (-1);
    return (atoi(got_from_read));
}

int main(int argc, char *argv[])
{
    /*int count=0; */
    if (argc < 3)
    {
        write(1,
              "usage: pidfilehack service /var/run/daemon.pid /usr/sbin/daemon args...\n",
              72);
        return 0;
    }
    if (unlink(argv[2]))
    {
        if (errno != ENOENT)
        {
            perror("could not remove pid file");
            return 1;
        }
    }

    /* try to fork */
    switch (fork())
    {

            /* if fail */
        case -1:
            perror("could not fork");
            return 2;

        case 0:                            /* child */
            /* The forks run the deamon */
            execve(argv[3], argv + 3, environ);
            perror("execvp failed");
            return 3;
    }

    sleep(1);
    /* try to load the pid file */
    do
    {
        int fd = 0;
        int len = 0;
        int ret = 0;
        int status = 0;
        char buf[21];
        char send[100];

        /* open pid file */
        fd = open(argv[2], O_RDONLY);

        /* if we succed to open */
        if (fd >= 0)
        {
            len = read(fd, buf, 20);
            close(fd);

            if (len > 0)
            {
                strcpy(send, "P");
                /* remove last newline */
                if (buf[len - 1] == '\n')
                    buf[len - 1] = 0;
                else
                    buf[len] = 0;

                strcat(send, buf);
                strcat(send, ":");
                strcat(send, argv[1]);

                /*printf("sending: \"%s\"\n", send); */
                ret = command_fetch_int(send);
                sleep(1);
                return (ret);
            }
            else
                printf("file there but open returned %d\n", fd);
        }                                   /* else printf("%s not there yet\n",argv[2]); */

        /* if children are dead, and did not get a pid, dont wait for a pid file */
        if (waitpid(-1, &status, WNOHANG) > 0)
            _exit(WEXITSTATUS(status));

        sleep(1);
        /* do this in an endless loop? */
        /*if (++count>=30)
         * _exit(1); */
    }
    while (1);
}
