/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

/* This file generates our services based on /etc/rcX.d */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <dirent.h>
#include "../../src/service_db.h"
#include "../../src/string_tools.h"
#include "../../src/initng_global.h"
#include "../../src/open_read_close.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"

#include "makerunlevel.h"

void clear_runlevel(char **runlevel)
{
    int i = 0;

    while (runlevel[i])
    {
        free(runlevel[i]);
        runlevel[i] = NULL;
        i++;
    }
}

service_h *make_new_service(char *runfile)
{
    service_h *new_service = NULL;

    service_db_new(&new_service, i_strdup(runfile), TYPE_SERVICE);
    char *script_start = NULL;
    char *script_stop = NULL;

    if (!new_service)
    {
        F_("Failed to alloc service %s\n", runfile);
        return (NULL);
    }

    /* Todo, this has to point at the same damn script, and be free'd once, no need to dup both start and stop blocks. */
    if (!open_read_close(runfile, &script_start))
    {
        F_("Failed to open %s\n", runfile);
        free(new_service->name);
        free(new_service);
        return (NULL);
    }
    if (!open_read_close(runfile, &script_stop))
    {
        F_("Failed to open %s\n", runfile);
        free(new_service->name);
        free(new_service);
        free(script_start);
        return (NULL);
    }

    service_db_set_string(START_SCRIPT, new_service, script_start);
    service_db_set_string(START_ARGS, new_service, i_strdup("start"));
    service_db_set_string(STOP_SCRIPT, new_service, script_stop);
    service_db_set_string(STOP_ARGS, new_service, i_strdup("stop"));


    return (new_service);
}

service_h *make_new_runlevel(char *runlevel_name)
{
    service_h *n_service;       /* service struct pointer too   */
    struct dirent **namelist;
    char **last_runlevel =
        (char **) initng_calloc(MAX_ENTRIES, sizeof(char *));
    char **current_runlevel =
        (char **) initng_calloc(MAX_ENTRIES, sizeof(char *));
    int current_runlevel_pos = 0;

    current_runlevel[0] = NULL;
    last_runlevel[0] = NULL;
    char **runlevel_tmp;
    service_h *new_service = NULL;
    int max, i, tmpi;
    char path[256];

    char tmp[2];

    int level = 0;
    int oldlevel = 0;
    char *name;
    char *runfile;

    sprintf(path, "/etc/%s.d", runlevel_name);

    /* allocate a new service */
    service_db_new(&n_service, i_strdup(runlevel_name), TYPE_RUNLEVEL);

    F_("make_new_runlevel(%s) check %s;\n", runlevel_name, path);

    max = scandir(path, &namelist, 0, alphasort);
    if (max < 0)
    {
        F_("Cant open %s!\n", path);
        if (n_service->name)
            free(n_service->name);
        if (n_service)
            free(n_service);
        if (namelist)
            free(namelist);
        return (NULL);
    }

    /* Main loop */
    for (i = 0; i < max; i++)
    {
        if (namelist[i]->d_name[0] == 'S')
        {
            /* Get current level of service (S*55*service) */
            strncpy(&tmp[0], &namelist[i]->d_name[1], 2);
            level = strtol(&tmp[0], (char **) NULL, 10);

            /* Get name of service */
            name = i_strdup(&namelist[i]->d_name[3]);
            runfile =
                calloc(strlen(name) + strlen("/etc/init.d/"), sizeof(char *));
            strcpy(runfile, "/etc/init.d/");
            strcat(runfile, name);
            D_("Adding service %s\n", runfile);

            /* Check if this service "level" is different, if it is we're on a new one. */
            /* Service level = S20Service, 20 is the level. */
            if (oldlevel != level)
            {
                runlevel_tmp = last_runlevel;
                last_runlevel = current_runlevel;
                current_runlevel = runlevel_tmp;
                clear_runlevel(current_runlevel);
                oldlevel = level;
                current_runlevel_pos = 0;
            }

            /* Do the actual adding of a service */
            new_service = make_new_service(runfile);
            if (!new_service)
            {
                F_("Failed to alloc service %s\n", name);
                free(name);
                free(runfile);
                continue;
            }

            /* We need to depend on the runlevel before. */
            if (strcmp(runlevel_name, "rcS") != 0)
                service_db_set_another_string(NEED, new_service,
                                              i_strdup("rcS"));


            //            printf("%s\n",runfile);
            //            printf("  Depends on ");
            /* Set dependencies to service "level" above us */
            tmpi = 0;
            while (last_runlevel[tmpi])
            {
                //                printf("%s ",last_runlevel[tmpi]);
                service_db_set_another_string(NEED, new_service,
                                              i_strdup(last_runlevel[tmpi]));
                tmpi++;
            }
            //            printf("\n");

            /* Add that service to "current_runlevel" so the next runlevel can depend on us */
            current_runlevel[current_runlevel_pos] = i_strdup(runfile);
            current_runlevel_pos++;

            /* We dont care if we fail here */
            service_db_set(DONE_WHEN_FAIL, new_service);
            service_db_set_string(FROM_FILE, new_service, i_strdup(runfile));

            /* Add service to service database */
            if (!service_db_add(new_service))
            {
                F_("can't add service to service db!\n");
                if (new_service->name)
                    free(new_service->name);
                if (new_service)
                    free(new_service);
                if (name)
                    free(name);
                if (runfile)
                    free(runfile);
                return (NULL);
            }

            /* Add so our custom runlevel depends on the service we just added */
            service_db_set_another_string(NEED, n_service, i_strdup(runfile));

            free(name);
            free(runfile);
        }
        free(namelist[i]);
    }
    free(namelist);

    service_db_set(UP_WHEN_DONE, n_service);
    service_db_set_string(FROM_FILE, n_service, i_strdup(path));

    printf("Returning from adding %s\n", runlevel_name);
    return (service_db_add(n_service));
}
