
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include "../../src/service_db.h"
#include "../../src/string_tools.h"
#include "../../src/initng_global.h"
#include "../../src/open_read_close.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"

#include "initng_rl_parser.h"

/* a simple parser for a runlevel file */
service_h *initng_rl_parser(char *runlevel_name)
{
    service_h *n_service;       /* service struct pointer too   */
    char filetoparse[200];
    char *w_depends = NULL;
    char *w = NULL;
    char *a = NULL;

    D_("initng_rl_parser(%s);\n", runlevel_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, runlevel_name);
    strcat(filetoparse, ".runlevel");
    D_("initng_rl_parser(%s): loading file %s.\n",
       runlevel_name, filetoparse);

    /* allocate a new service */
    service_db_new(&n_service, i_strdup(runlevel_name), TYPE_RUNLEVEL);

    /* copy file filename into allocated file_buf */
    if (!open_read_close(filetoparse, &w_depends))
    {
        D_("parse_file(%s): Cant open config file!\n", filetoparse);
        if (n_service->name)
            free(n_service->name);
        if (n_service)
            free(n_service);
        return (NULL);
    }
    w = w_depends;

    st_jump_to_valid_line(&w);
    while (w[0] != '\0')
    {
        a = st_dup_line(&w);
        D_("adding dep: \"%s\"\n", a);
        service_db_set_another_string(NEED, n_service, a);
        st_jump_to_valid_line(&w);
    }

    free(w_depends);

    service_db_set(UP_WHEN_DONE, n_service);
    service_db_set_string(FROM_FILE, n_service, i_strdup(filetoparse));

    return (service_db_add(n_service));
}

int module_init(void)
{

    D_("initng_rl_parser: moule_init();\n");
    return (initng_add_hook(PARSERS, &initng_rl_parser));
}

void module_unload(void)
{
    initng_del_hook(PARSERS, &initng_rl_parser);
}
