
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <stdio.h>
#include <string.h>                         /* strdup() */
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>

#include "../../src/active_db.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_execute.h"
#include "../../src/string_tools.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_common.h"
#include "../../src/initng_signal.h"
/*
   #include "initng_global.h"
   #include "process_db.h"
   #include "main.h"
   #include "string_tools.h"
   #include "initng_execute.h"
   #include "bash_executor.h"
   #include "active_db.h"
   #include "initng_load_module.h"
   #include "initng_toolbox.h" */

#include "initng_simple_launcher.h"

static pid_t
initng_execute_fork_and_exec(process_h * process_to_exec, active_h * s,
                             char *e, char *ea, char *pf);

int initng_s_start(active_h * service)
{
    char *e = NULL;
    char *t = NULL;

    assert(service);
    assert(service->from_service);
    /* WE ARE EXECUTING A START FILE */
    if (active_db_is(START, service))
    {
        while ((e = active_db_get_next_string(START, service, e)))
        {

            struct stat test;

            t = fix_variables(e, service);
            if (stat(t, &test) == -1)
            {
                F_("Stat failed on %s\n", t);
                free(t);
                continue;

            }


            if (initng_execute_fork_and_exec
                (service->start_process, service, t,
                 active_db_get_string(START_ARGS, service),
                 active_db_get_string(PID_FILE, service)))
            {
                free(t);
                return (TRUE);
            }
            D_("%s did not work\n", t);
            free(t);
        }
        return (FAIL);
    }

    if (active_db_is(DAEMON, service))
    {
        while ((e = active_db_get_next_string(DAEMON, service, e)))
        {

            struct stat test;

            t = fix_variables(e, service);

            if (stat(t, &test) == -1)
            {
                D_("Stat failed on start %s, file dont exits, trying next if any.\n", t);
                free(t);
                continue;
            }

            if (initng_execute_fork_and_exec
                (service->start_process, service, t,
                 active_db_get_string(DAEMON_ARGS, service),
                 active_db_get_string(PID_FILE, service)))
            {
                free(t);
                return (TRUE);
            }

            D_("start %s did not work, trying next if any.\n", t);
            free(t);
        }
        return (FAIL);

    }
    return (FALSE);
}

int initng_s_stop(active_h * service)
{
    char *e = NULL;
    char *t = NULL;

    assert(service);
    assert(service->from_service);


    /* WE ARE EXECUTING A START SCRIPT */
    if (active_db_is(STOP, service))
    {
        while ((e = active_db_get_next_string(STOP, service, e)))
        {

            struct stat test;

            t = fix_variables(e, service);

            if (stat(t, &test) == -1)
            {
                D_("Stat failed on stop %s\n", t);
                free(t);
                continue;
            }


            if (initng_execute_fork_and_exec
                (service->stop_process, service, t,
                 active_db_get_string(STOP_ARGS, service),
                 active_db_get_string(PID_FILE, service)))
            {
                free(t);
                return (TRUE);
            }
            D_("stop %s did not work\n", t);
            free(t);
        }
        return (FAIL);

    }
    return (FALSE);
}


static pid_t
initng_execute_fork_and_exec(process_h * process_to_exec, active_h * s,
                             char *e, char *ea, char *pf)
{

    /* called from inside the service directory, return the PID or 0 on error */

    /* This is the real service kicker */
    pid_t pid_fork;             /* pid got from fork() */
    int try_count = 0;          /* Count tries */
    char **argv = 0;            /* Arguments */
    int argc;                   /* number of arguments */

    assert(process_to_exec);
    assert(e);

    /* Try to fork 3 times */
    while ((pid_fork = fork()) == -1)
    {
        if (try_count > 3)
            return 0;                       /* Already tried 3 times = no more try */
        sleep(++try_count * 2);             /* Sleep a while before trying agaim */
    }

    /* if this is child */
    if (pid_fork == 0)
    {
        char *exec = NULL;
        char *exec_args = NULL;
        char *tmp = NULL;
        int t = 0;

        disable_signals();

        ea = fix_variables(ea, s);
        pf = fix_variables(pf, s);

        if (active_db_is(NICE, s))
        {
            nice(active_db_get_int(NICE, s));
        }


        /* P I D F I L E   H A C K */
        /* if we have to take pid form a pidfile, use pidfilehack */
        if (pf)
        {
            if (s_is(I_AM_INIT))
                exec = i_strdup("/sbin/pidfilehack");
            else
                exec = i_strdup("/sbin/pidfiledhack");

            /*  pidfilehack <service> <pidfile> <daemon> <deamon_args> */
            exec_args = initng_calloc(1, 200 * sizeof(char));
            strcpy(exec_args, s->name);
            strcat(exec_args, " ");
            strcat(exec_args, pf);
            strcat(exec_args, " ");
            strcat(exec_args, e);
            if (ea)
            {
                strcat(exec_args, " ");
                strcat(exec_args, ea);
            }
            free(ea);
            free(e);
            D_("FROM_FORK initng_execute_fork_and_exec(%s) : pidfilehack : (%s,%s);\n", s->name, exec, exec_args);
        }
        else
        {
            exec = e;
            exec_args = ea;
        }

        D_("FROM_FORK initng_execute_fork_and_exec(%s,%s);\n", e, ea);

        if (s_is(I_AM_INIT))
        {
            ioctl(0, TIOCNOTTY, 0);
            setsid();                       /* Run a program in a new session ??? */
        }

        set_pipes(process_to_exec);

        if (s_is(I_AM_INIT))
            tcsetpgrp(0, getpgrp());        /* run this in foreground on fd 0 */

        if ((tmp = active_db_get_string(CHDIR, s)))
        {
            D_("change dir to: %s\n", tmp);
            chdir(tmp);
        }


        if ((tmp = active_db_get_string(CHROOT, s)))
        {
            D_("change root to: %s\n", tmp);
            chdir(tmp);
            chroot(tmp);
        }

        if ((tmp = active_db_get_string(SUID, s)))
        {
            struct passwd *pw;

            pw = getpwnam(tmp);
            if (!pw)
                F_("%s: USER %s not found!", s->name, tmp);
            D_("setuid to: %i.\n", pw->pw_uid);
            setuid(pw->pw_uid);
        }

        /* get args from service database */
        if (exec_args)
        {
            D_("FROM_FORK initng_execute_fork_and_exec(%s): start args: %s\n",
               exec, exec_args);

            /* if we only vfork, we need to copy this string because split destroys it */
            /*temp=i_strdup(process_to_exec->exec_args); */
            /* split args */
            argv = split_n(exec_args, &argc, 2, 1);

            /* i dont know about this */
            if (!argv[argc - 1])
            {
                argv[argc - 1] = '\0';      /* the last */
            }
            else
            {
                argv[argc] = '\0';          /* must be null */
            }
        }
        else
        {
            argv = initng_calloc(2, sizeof(char *));    /* malloc a null */
            argv[1] = '\0';                 /* set to null */
        }

        /* set argv[0] */
        argv[0] = exec;


        /* if stdin not points to standard */
        /*        if (process_to_exec->stdin != 0)
           {
           dup2(process_to_exec->stdin, 0);     copy filedescriptor 
           fcntl(0, F_SETFD, 0);
           }

           and if stdout not points to usual 
           if (process_to_exec->stdout != 1)
           {
           dup2(process_to_exec->stdout, 1);    copy out stdout 
           dup2(process_to_exec->stdout, 2);    and stderr 
           fcntl(1, F_SETFD, 0);
           fcntl(2, F_SETFD, 0);
           } */


        {
            int i;

            for (i = 0; argv[i]; i++)
                D_("i.e.f.a.e.(): argv[%i]=\"%s\"\n", i, argv[i]);
        }



        /* delay execution a bit */
        if ((t = active_db_get_int(UDELAY, s)) > 0)
        {
            D_("Sleeping for %i miliseconds.\n", t);
            usleep(t);
        }
        else if ((t = active_db_get_int(DELAY, s)) > 0)
        {
            D_("Sleeping for %i seconds.\n", t);
            sleep(t);
        }
        else
        {
            /* do a minimum sleep, to let the mother process 
               to register child, and notice death */
            usleep(ALL_USLEEP);
        }

        /* FINALLY EXECUTE *//* execve replaces the running process */
        execve(argv[0], argv, new_environ(s));

        /* Will never get here if execve succeeded */
        F_("EEEEEEEEEEEEEEEEEEERRRRRRRRRRRRRRRRRRRRRRRRRRROOOOOOOOOOOOOOOOOOOOOOOOOORRRRRRRRRRRRRRR!!!!!!!!!\n");
        F_("Cant execute source %s, \"%s\" !\n", e, ea);

        /* free variable, good for making valgrind quit */
        free(argv[0]);
        free(argv);
        free(exec_args);

        /* clean up */
        initng_free();

        /* exit 1, why???, to emit signal? */
        _exit(1);
    }

    /* save pid of fork */
    D_("FROM_FORK Forkstarted pid %i.\n", pid_fork);
    process_to_exec->pid = pid_fork;
    return pid_fork;

    /* if to test want to lock this up util fork is done ...   waitpid(pid_fork,0,0); */
}                                           /* end fork_and_exec() */


int module_init(void)
{
    D_("initng_simple_plugin: module_init();\n");

    initng_add_hook(STARTL, &initng_s_start);
    initng_add_hook(STOPL, &initng_s_stop);

    return (TRUE);
}

void module_unload(void)
{
    D_("initng_simple_plugin: module_unload();\n");

    initng_del_hook(STARTL, &initng_s_start);
    initng_del_hook(STOPL, &initng_s_stop);

}
