
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */


#include "../../src/initng.h"

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/active_db.h"
#include "../../src/initng_history.h"
#include "../../src/main.h"
#include "../../src/initng_toolbox.h"
#include "../../src/initng_plugin_hook.h"

#include "libsplash/libsplash.h"

#include "initng_splash_out.h"

#define SPLASH_UTIL_PATH "/sbin/splash_util"
#define SPLASH_UTIL_STATIC_PATH "/sbin/splash_util.static"
#define SPLASH_UTIL_THEME "--theme=default"

int do_splash = 1;
char *splash_util = NULL;
extern char **environ;

/* THIS IS MAIN */
void splash_progress(int p)
{
    /* check if do_splash is true. */
    if (!do_splash)
        return;
    if (fork())
    {
        D_("splash_progress(%i): This is not fork.\n", p);
        return;
    }
    else
    {

#ifdef EXECSPLASH
        char **arg = NULL;
        struct stat f_stat;
        arg = (char **) initng_calloc(7, sizeof(char *));
        arg[0] = splash_util;
        arg[1] = "--vc=0";
        arg[2] = (char *) initng_calloc(20, sizeof(char));
        arg[3] = SPLASH_UTIL_THEME;
        arg[4] = "--mode=s";
        arg[5] = "--cmd=paint";
        arg[6] = NULL;
        sprintf(arg[2], "--progress=%i", p * 655);

        /* Probe that fb exists */
        if (!
            (stat("/dev/fb0", &f_stat) == 0
             || stat("/dev/fb/0", &f_stat) == 0))
        {
            D_("splash_out.so.0.0: Can't temporary find framebuffer on /dev/fb0 or /dev/fb/0\n");
            return;
        }

        D_("splash_progress(%i): %s %s %s %s %s %s\n", p, arg[0], arg[1],
           arg[2], arg[3], arg[4], arg[5]);
        execve(splash_util, arg, environ);
        F_("splash_progress(%i): Failed executing: %s %s %s %s %s %s\n", p,
           arg[0], arg[1], arg[2], arg[3], arg[4], arg[5]);
        _exit(1);                           /* will not get here */
#else
        set_progress(SPLASH_UTIL_THEME, p * 655);
        _exit(0);
#endif
    }

    return;
}

void splash_remove(void)
{
    do_splash = 0;
    if (fork())
    {
        D_("splash_remove(): This is not fork.\n");
        return;
    }
    else
    {
        char *arg[] = { splash_util, "--cmd=setmode", "--mode=v", NULL };
        D_("splash_remove(): %s %s %s\n", arg[0], arg[1], arg[2]);
        execve(splash_util, arg, environ);
        D_("splash_remove(): Failed executing: %s %s %s\n", arg[0], arg[1],
           arg[2]);
        _exit(1);
    }
    return;
}


void splash_service(active_h * service)
{
    if (!service)
        return;

    /* if, do some output and history adds */
    switch (service->a_status)
    {

        case STARTING:
            if (g.sys_state == STATE_STARTING)
                splash_progress(active_db_percent_started());
            return;
        case RUNNING:
            if (g.sys_state == STATE_STARTING)
                splash_progress(active_db_percent_started());
            return;
        case START_DEP_MET:
            /* if deps for X are up, unload and stop splash */
            /* dont do cpuintensive strcmp if first char not is a 'd' */
            if (service->name && service->name[0] == 'd')
                if (strcmp("daemon/gdm", service->name) == 0 ||
                    strcmp("daemon/kdm", service->name) == 0 ||
                    strcmp("daemon/wdm", service->name) == 0 ||
                    strcmp("daemon/xdm", service->name) == 0)
                {
                    /* remove splash image */
                    splash_remove();
                }
        default:
            return;
    }
}


static void splash_system(h_sys_state state)
{
    switch (state)
    {
        case STATE_UP:
            splash_remove();
            break;
        default:
            break;
    }
}

int module_init(void)
{
    struct stat f_stat;

    D_("moule_init();\n");

    /* Probe that fb exists */
    if (!(stat("/dev/fb0", &f_stat) == 0 || stat("/dev/fb/0", &f_stat) == 0))
    {
        F_("splash_out.so.0.0: Can't find framebuffer on /dev/fb0 or /dev/fb/0\n");
        return (FALSE);
    }


    if (stat(SPLASH_UTIL_STATIC_PATH, &f_stat) != 0)
    {
        if (stat(SPLASH_UTIL_PATH, &f_stat) != 0)
        {
            F_("splash_out.so.0.0: Cant find " SPLASH_UTIL_PATH ".\n");
            return (FALSE);                 /* don't load handlers */
        }
        else
        {
            splash_util = i_strdup(SPLASH_UTIL_PATH);
        }
    }
    else
    {
        splash_util = i_strdup(SPLASH_UTIL_STATIC_PATH);
    }


    /* add initng_colorprint_out to list of service state watchers */
    initng_add_hook(WATCHERS, &splash_service);

    /* add initng_colorprint_out to list of system state watchers */
    initng_add_hook(SWATCHERS, &splash_system);

    return (TRUE);
}

void module_unload(void)
{
    D_("module_unload();\n");

    /* remove splash image */
    splash_remove();

    /* del initng_colorprint_out from list of service state watchers */
    initng_del_hook(WATCHERS, &splash_service);

    /* del initng_colorprint_out from list of system state watchers */
    initng_del_hook(SWATCHERS, &splash_system);

}
