/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 neuron <neuron@hollowtube.mine.nu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "../../src/initng.h"


#include <stdio.h>
#include <stdlib.h>                         /* malloc() free() exit() */
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#include "../../src/initng_handler.h"
#include "../../src/initng_global.h"
#include "../../src/initng_plugin_hook.h"
#include "../../src/initng_common.h"
#include "../../src/initng_toolbox.h"

#include "initng_stdout.h"

local_plugin_database_h *find_service(active_h * service)
{
    /* Scans through our local database, returns a new one if we cant find an old one. */
    int i;
    int foundfree = 0;

    D_("Scan for service %s\n", service->name);
    for (i = 0; i < MAX_ACTIVE; i++)
        if (local_database[i])
        {
            if (strcmp(local_database[i]->service->name, service->name) == 0)
                return local_database[i];
        }
        else
            foundfree = i;

    /* No free service found */
    if (foundfree == 0)
    {
        F_("Service database full!");
        return (FALSE);
    }
    D_("Allocating new service\n");
    local_database[foundfree] =
        (local_plugin_database_h *) initng_calloc(1,
                                                  sizeof
                                                  (local_plugin_database_h));
    local_database[foundfree]->service = service;
    local_database[foundfree]->fd = 0;
    local_database[foundfree]->append = FALSE;
    return local_database[foundfree];
}

void program_output(active_h * service, int datalen, int forceflush)
{
    process_h *x;
    char *filename = NULL;

    if (!active_db_is(stdout_id, service))
        return;

    if (service->start_process)
    {
        x = service->start_process;
    }
    else if (service->stop_process)
    {
        x = service->stop_process;
    }
    else
        return;

    local_plugin_database_h *entry = find_service(service);

    if (entry == NULL)
        return;

    if (entry->fd == 0)
    {
        filename =
            fix_variables(active_db_get_string(stdout_id, service), service);
        if (!entry->append)
            entry->fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC);
        else
            entry->fd = open(filename, O_WRONLY | O_CREAT | O_APPEND);
        if (entry->fd < 1)
        {
            entry->fd = 0;
            F_("Error opening %s, err : %s\n", filename, strerror(errno));
        }
    }
    D_("Writing data...\n");
    if (write(entry->fd, &x->buffer[x->buffer_pos], datalen) != datalen)
        F_("Error writing to %s's log, err : %s\n", service->name,
           strerror(errno));
    free(filename);

    if (forceflush)
        if (entry->fd > 0)
        {
            entry->append = TRUE;
            close(entry->fd);
            entry->fd = 0;
            entry->service = NULL;
        }
}

void system_state(h_sys_state state)
{
    if (state == STATE_ASE)
        module_unload();
}

int module_init(void)
{
    D_("moule_init();\n");
    int i;

    stdout_id = initng_sdt_add(STRING, "stdout", FALSE);
    local_database =
        initng_calloc(MAX_ACTIVE + 1, sizeof(local_plugin_database_h));
    for (i = 0; i < MAX_ACTIVE; i++)
        local_database[i] = NULL;


    initng_add_hook(SWATCHERS, &system_state);
    initng_add_hook(PIPEWATCHERS, &program_output);
    return (TRUE);
}

void module_unload(void)
{
    int i;

    D_("module_unload();\n");

    for (i = 0; i < MAX_ACTIVE; i++)
        if (local_database[i] && (local_database[i]->fd > 0))
            close(local_database[i]->fd);

    for (i = 0; i < MAX_ACTIVE; i++)
        if (local_database[i])
            free(local_database[i]);
    free(local_database);
    local_database = NULL;

    initng_del_hook(SWATCHERS, &system_state);
    initng_del_hook(PIPEWATCHERS, &program_output);
}
