
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <unistd.h>                         /* execv() pipe() usleep() pause() chown() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <stdio.h>
#include <string.h>                         /* strdup() */
#include <assert.h>
#include <pwd.h>
#include <sys/stat.h>

#include "initng_global.h"
#include "process_db.h"
#include "main.h"
#include "string_tools.h"                   /* split() */
#include "initng_execute.h"
#include "active_db.h"
#include "initng_plugin_callers.h"
#include "initng_toolbox.h"
#include "initng_common.h"

int start_launch(active_h * service)
{
    assert(service);
    assert(service->from_service);


    D_("start_launch(%s);\n", service->name);


    process_db_new(&service->start_process, T_START);
    if (!service->start_process)
    {
        F_("start_launch(%s): unable to allocate start_process!\n",
           service->name);
        return (FALSE);
    }

    switch (initng_plugin_launch_start(service))
    {
        case FAIL:
            F_("initng_execute(%s): FAIL EXECUTING START\n", service->name);
            process_db_free(&service->start_process);
            return (FAIL);
        case FALSE:                        /* done direct */
            process_db_free(&service->start_process);
            return (FALSE);
    }

    return (TRUE);
}


int stop_launch(active_h * service)
{
    assert(service);
    assert(service->from_service);

    D_("stop_launch(%s);\n", service->name);

    /* create a new stop process */
    process_db_new(&service->stop_process, T_STOP);
    if (!service->stop_process)
    {
        F_("stop_launch(%s): unable to allocate stop_process!\n",
           service->name);
        return (FALSE);
    }

    switch (initng_plugin_launch_stop(service))
    {
        case FAIL:
            F_("initng_execute(%s): FAIL EXECUTING STOP\n", service->name);
            process_db_free(&service->stop_process);
            return (FAIL);
        case FALSE:                        /* done direct */
            process_db_free(&service->stop_process);
            return (FALSE);
    }

    return (TRUE);
}


/* creates an custom environment variable */
char **new_environ(active_h * s)
{
    int nr = 0;
    char **env;

    /* Better safe than sorry... */
    if (s && s->name == NULL)
        s->name = "unknown";

    /* count existing env,s */
#ifdef USE_OLD_ENV
    while (environ[nr])
        nr++;
#endif
#ifdef DEBUG
    if (s)
        D_("new_environ(): there are %i envs in config %s!\n",
           active_db_count_type(ENV, s), s->name);
#endif
    /*  Add 11 more environs */
    if (s)
        nr = nr + 11 + active_db_count_type(ENV, s);

    /* allocate */
    env = initng_calloc(nr, sizeof(char *));

    /* duplicate */
#ifdef USE_OLD_ENV
    for (nr = 0; environ[nr]; nr++)
    {
        env[nr] = i_strdup(environ[nr]);
    }
#else
    nr = 0;
#endif
    /* add costum */
    env[nr++] = i_strdup("INITNG=" INITNG_VERSION);
    env[nr++] =
        i_strdup
        ("PATH=/bin:/usr/bin:/sbin:/usr/sbin:/usr/local/bin:/opt/bin");
    env[nr++] = i_strdup("HOME=/root");
    env[nr++] = i_strdup("USER=root");
    env[nr++] = i_strdup("TERM=linux");
    /*env[nr++] = i_strdup("BASH_ENV=\"" INITNG_ROOT "/bash-env.sh\""); */

    if (s)
    {
        assert(s);
        env[nr] =
            (char *) initng_calloc(1, sizeof(char) * (9 + strlen(s->name)));
        strcpy(env[nr], "SERVICE=");
        strcat(env[nr], s->name);
        nr++;

        env[nr] =
            (char *) initng_calloc(1, sizeof(char) * (6 + strlen(s->name)));
        strcpy(env[nr], "NAME=");
        strcat(env[nr], st_strip_path(s->name));
        nr++;

        env[nr] =
            (char *) initng_calloc(1, sizeof(char) * (7 + strlen(s->name)));
        strcpy(env[nr], "CLASS=");
        strcat(env[nr], s->name);
        st_strip_end(&env[nr]);
        nr++;


        env[nr] = NULL;
        /* insert all env strings form config */
        if ((env[nr] = active_db_get_next_string(ENV, s, NULL)))
        {
            nr++;
            while ((env[nr] = active_db_get_next_string(ENV, s, env[nr - 1])))
                nr++;

        }
    }
    /* null last */
    if (env[nr] != NULL)
        env[nr++] = NULL;

#ifdef DEBUG
    for (nr = 0; env[nr]; nr++)
    {
        D_("new_environ(): environ[%i] = \"%s\"\n", nr, env[nr]);
    }
#endif

    /* return new environ */
    return (env);
}

/* this frees an environment variable */
void free_environ(char **tf)
{
    int i = 0;

    D_("free_environ();\n");
    assert(tf);
    for (i = 0; tf[i]; tf++)
    {
        free(tf[i]);
    }
    free(tf);
}

/* Set pipes for a process THIS FUNCTION SHOULD BE CALLED IN A FORK, it closes stdout[0] */
void set_pipes(process_h * process)
{
    /* close stdin/stdout/stderr */
    close(STDOUT_FILENO);
    close(STDERR_FILENO);

    /* Dup them all to the correct pipes */

    dup2(process->stdout[1], STDOUT_FILENO);
    dup2(process->stdout[1], STDERR_FILENO);

    /* TODO : no idea what this does :p, but it was there before, is this same as close() on top? */
    fcntl(STDIN_FILENO, F_SETFD, 0);
    fcntl(STDERR_FILENO, F_SETFD, 0);

    /* Close the sides of the pipes we dont need, as we're in fork we wont need this part. */
    close(process->stdout[0]);
    process->stdout[0] = -1;
}

/* open out dev for service output */
void open_dev_out(void)
{
    struct stat st;
    int fd1 = 1;
    int fd2 = 2;
    char *s_stdout = STD_OUT;
    char *s_stderr = STD_ERR;



    /* set defaults if not entered */
    if (stat(s_stdout, &st) != 0)
    {
        s_stdout = STD_OUT2;
        if (stat(s_stdout, &st) != 0)
        {
            s_stdout = "/dev/console";
            if (stat(s_stdout, &st) != 0)
            {
                F_("Cant find /dev/console!\n");
                return;
            }
        }
    }

    if (stat(s_stderr, &st) != 0)
    {
        s_stderr = "/dev/console";
        if (stat(s_stderr, &st) != 0)
        {
            F_("Cant find /dev/console!\n");
            return;
        }
    }

    /* if both stdout and stderr are the same file */
    if (strcasecmp(s_stdout, s_stderr) == 0)
    {
        D_("open_dev_out(): Will open %s for stdout.\n", s_stdout);
        fd1 = open(s_stdout, O_RDWR | O_NOCTTY | O_CREAT | O_APPEND);
        if (fd1)
            dup2(fd1, fd2);
    }
    else
    {
        D_("open_dev_out(): Will open %s for stdout and %s for stderr",
           s_stdout, s_stderr);
        fd1 = open(s_stdout, O_RDWR | O_NOCTTY | O_CREAT | O_APPEND);
        fd2 = open(s_stderr, O_RDWR | O_NOCTTY | O_CREAT | O_APPEND);
    }

    if (fd1 > 2)
    {
        /* what are we gonna do with input? */
        dup2(fd1, 0);
        dup2(fd1, 1);
        close(fd1);
    }
    else
    {
        F_(" Unable to set out pipe 1> !\n");
        return;
    }
    if (fd2 > 2)
    {
        dup2(fd2, 2);
        close(fd2);
    }
    else
    {
        F_(" Unable to set out pipe 2> !\n");
        return;
    }

}
