
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include "initng_history.h"
#include "active_status.h"
#include "initng_global.h"
#include "initng_toolbox.h"
#include <sys/types.h>                      /* time_t */
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <stdio.h>                          /* printf() */
#include <assert.h>

                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                           /*history_h **history_db = NULL; *//* a big list of pointers pointing to all known history structs */
                                                                                                                                                                                                                                                                                                                                                                                      /*int next_history = 0; *//* a counter marking where to write next entry in big history database */

static void history_free_this(history_h ** this_one);

/* clear history */
void history_db_init(void)
{
#ifdef ZERO_MEMORY
    int i;                      /* counter */
#endif
    g.history_db =
        (history_h **) initng_calloc(HISTORY + 1, sizeof(history_h));
    g.next_history = 0;
    /* Null history */

#ifdef ZERO_MEMORY
    /* when using initng_calloc, variables is zeroed  */
    for (i = 0; i <= HISTORY; i++)
    {
        g.history_db[i] = NULL;
    }
#endif
}

void history_db_compensate_time(time_t skew)
{
    int i = 0;                  /* counter */

    D_("history_db_compensate_time(%i);\n", (int) skew);

    /* change time */
    for (i = 0; i <= HISTORY; i++)
        if (g.history_db[i] && g.history_db[i]->time)
            g.history_db[i]->time = g.history_db[i]->time + skew;
}

/* clear history */
void history_free_all(void)
{
    int i = 0;                  /* counter */

    /* Null history */
    for (i = 0; i <= HISTORY; i++)
    {
        if (g.history_db[i])
        {
            if (g.history_db[i]->name)
                free(g.history_db[i]->name);
            free(g.history_db[i]);
        }
    }
    free(g.history_db);
}

/* sanely clear and frees a history struct */
static void history_free_this(history_h ** this_one)
{
    /*int i=0; */
    /* dumb check */
    if (*this_one)

        /* free historyname if any */
        if ((*this_one)->name)
            free((*this_one)->name);

    /* free history_struct */
    free((*this_one));
    (*this_one) = NULL;
}

/* add values to history database */
void history_add_values(char *name, time_t time, e_a_status action)
{
    history_h *tmp_e = NULL;    /*temporary pointer to insert data and add to db */

    assert(name);

    /* dumb check */
    if (!time)
        return;

    D_("adding: %s.\n", name);
    /* allocate space for data */
    if (!(tmp_e = initng_calloc(1, sizeof(history_h))))
    {
        F_("history_add_values(): Out of memory.\n");
        return;
    }

    /* set data in struct */
    tmp_e->name = i_strdup(name);
    tmp_e->time = time;
    tmp_e->action = action;

    /* add struct */
    history_add(tmp_e);

    /* leave */
    return;
}

/* add to history space */
void history_add(history_h * entry)
{
    /* dumb check */
    assert(entry);

    /* add pointer */
    g.history_db[g.next_history] = entry;

    D_("adding: %s at %i.\n", entry->name, g.next_history);

    /* step forward the next counter */
    if (g.next_history == HISTORY)
        g.next_history = 0;
    else
        g.next_history++;

    /* Make sure next is free so we don't get a rolling loop */
    history_free_this(&g.history_db[g.next_history]);
    g.history_db[g.next_history] = NULL;    /* to be sure, TODO Delete this?? */

    /* exit */
    return;
}

/* fetch path to next history, or first if got none */
history_h *history_get_next(history_h * last)
{
    int i;                      /* counter */

    D_("next_history:%i\n", g.next_history);

    /* if not, give first */
    if (last)
    {
        /* if we got last, step forward i to it */
        for (i = 0; i <= HISTORY; i++)
        {
            if (g.history_db[i] == last)
            {                               /* if this is what we are looking for */
                break;
            }
        }

        if (i > HISTORY)
            i = 0;
        else
            i++;
    }
    else
    {
        /* set first history to next history */
        i = g.next_history + 1;
    }

    /* Try every one but the empty last one the next_history to allocate */
    while (i != g.next_history)
    {
        /* if found some try return next */
        if (g.history_db[i])
        {
            D_("returning [%i]:%s\n", i, g.history_db[i]->name);
            return (g.history_db[i]);
        }

        /* step forward or around */
        if (i >= HISTORY)
        {
            i = 0;                          /* start over */
        }
        else
        {
            i++;                            /* step forward */
        }
    }
    /* return null - We did not find anything */
    return (NULL);
}

/* returns the time in seconds, sense history entry name, switched to status from */
int time_been_from(char *name, e_a_status from)
{
    history_h *entry = NULL;
    time_t from_t = 0;
    time_t time_now = time(0);

    assert(name);

    while ((entry = history_get_next(entry)))
    {
        if (strcmp(entry->name, name) == 0)
        {
            if (entry->action == from)
                from_t = entry->time;
        }
    }

    /* if now time is greater, more then last time (???) */

    if (from_t && time_now > from_t)
        return ((int) time_now - (int) from_t);
    else
        return (0);

}
