#include <stdlib.h>
#include <sys/stat.h>
#include <stdio.h>
#include <dlfcn.h>
#include <assert.h>
#include <string.h>
#include "initng.h"
#include "service_db.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_global.h"

/* this will null all arrays */
void initng_load_module_init(void)
{
    int i;

    for (i = 0; i < MAX_MODULES; i++)
    {
        g.modules[i].module_name = NULL;
        g.modules[i].module = NULL;
    }

    for (i = 0; i < MAX_CALLERS; i++)
        g.callers[i].ptype = NOPLUG;
}

m_h *initng_get_next_module(m_h * c)
{
    int i = 0;

    /* fast forward */
    if (c)
        for (; i < MAX_MODULES; i++)
            if (&g.modules[i] == c)
            {
                i++;
                break;
            }

    for (; i < MAX_MODULES; i++)
        if (g.modules[i].module_name)
            return (&g.modules[i]);

    return (NULL);
}

/* load a module= a parser module for now */
int initng_load_module(char *module_name)
{
    int i;
    int ret = 0;
    char *error = NULL;
    int (*init_mod) (void);
    struct stat st;

    assert(module_name);

    /* check that file exists */
    if (stat(module_name, &st) != 0 || S_ISREG(st.st_mode) != 1)
    {
        F_("File \"%s\" not found, or not a regular file!\n\n", module_name);
        return (FALSE);
    }

    /* Try find duplets */
    for (i = 0; i < MAX_MODULES; i++)
        if (g.modules[i].module_name
            && strcmp(g.modules[i].module_name, module_name) == 0)
        {
            F_("Module \"%s\" alredy loaded, wont lode it twice!\n\n",
               module_name);
            return (FALSE);
        }


    /* find a free entry */
    for (i = 0; i < MAX_MODULES; i++)
        if (!g.modules[i].module)
            break;

    /* open module */
    g.modules[i].module = dlopen(module_name, RTLD_LAZY);

    if (!g.modules[i].module || (error = dlerror()))
    {
        if (error)
        {
            F_("%s\n\n", error);
            free(error);
        }
        else
            F_("Problem loading %s\n", module_name);
        return (FALSE);
    }

    D_("Succeeded to open lib \"%s\"\n\n", module_name);

    init_mod = dlsym(g.modules[i].module, "module_init");

    if ((error = dlerror()))
    {
        F_("dlerror: %s\n\n", error);
    }

    /* set module name in database */
    g.modules[i].module_name = i_strdup(module_name);

    if (init_mod)
        ret = (*init_mod) ();
    D_("module_init() for module %s return: %i\n", g.modules[i].module_name,
       ret);

    if (ret < 1)
    {
        F_("module %s did not load corrrecly (ret code %i), unloading in 1 sec...\n", g.modules[i].module_name, ret);
        sleep(1);
        initng_unload_module(&g.modules[i]);
    }

    return (TRUE);
}

int initng_unload_module(m_h * module)
{
    char *error;
    void *(*unload_mod) (void);

    assert(module);
    D_("initng_unload_module(): Unloading: %s.\n", module->module_name);

    unload_mod = dlsym(module->module, "module_unload");

    if ((error = dlerror()) != NULL)
    {
        F_("dlerror: %s\n\n", error);
    }

    if (unload_mod)
        (*unload_mod) ();

    dlclose(module->module);
    module->module = NULL;
    if (module->module_name)
        free(module->module_name);
    module->module_name = NULL;
    return (TRUE);
}

/* this will dlclose all arrays */
void initng_load_module_close_all(void)
{
    int i;

    D_("initng_load_module_close_all(void);\n");

    for (i = 0; i < MAX_MODULES; i++)
        if (g.modules[i].module)
        {
            initng_unload_module(&g.modules[i]);
        }
}
