#include <stdlib.h>
#include <stdio.h>
#include <dlfcn.h>
#include <assert.h>
#include "initng.h"
#include "service_db.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_global.h"


#include <fcntl.h>                          /* fcntl() */
#include <errno.h>
#include <string.h>


/* P L U G I N    C A L L E R S */

int initng_plugin_launch_start(active_h * s)
{
    int i;
    int to_ret;

    for (i = 0; i < MAX_CALLERS; i++)
    {
        if (g.callers[i].ptype == STARTL)
        {
            if (!(to_ret = (*g.callers[i].c.launcher) (s)) == FALSE)
            {
                return (to_ret);
            }
        }
    }
    return (FALSE);
}

int initng_plugin_launch_stop(active_h * s)
{
    int i;
    int to_ret;

    for (i = 0; i < MAX_CALLERS; i++)
    {
        if (g.callers[i].ptype == STOPL)
        {
            if (!(to_ret = (*g.callers[i].c.launcher) (s)) == FALSE)
            {
                return (to_ret);
            }
        }
    }
    return (FALSE);
}

void initng_plugin_main(void)
{
    int i;

    for (i = 0; i < MAX_CALLERS; i++)
        if (g.callers[i].ptype == MAIN)
            (*g.callers[i].c.main) ();
}

void initng_close_all_fd(void)
{
    int i;

    for (i = 0; i < MAX_CALLERS; i++)
    {
        if (g.callers[i].ptype == FDWATCHERS && (g.callers[i].c.fdh.fds > 0))
        {
            close(g.callers[i].c.fdh.fds);
            g.callers[i].c.fdh.fds = 0;
            g.callers[i].c.fdh.call_module = NULL;
            g.callers[i].ptype = NOPLUG;

        }
    }

}

/* Calls plugins when data is ready in pipe */
void initng_plugin_readpipe(active_h * service, int datalen, int force_flush)
{
    int i;

    for (i = 0; i < MAX_CALLERS; i++)
    {
        if (g.callers[i].ptype == PIPEWATCHERS)
        {
            (*g.callers[i].c.pipewatcher) (service, datalen, force_flush);
        }
    }
    return;
}

/* This does the actual read from pipe, all checks have been done here */
int initng_process_readpipe_read(active_h * service, process_h * x,
                                 int program_end)
{
    int forceflush = program_end;
    int maxread = MAX_BUFFER - x->buffer_pos;
    char *buffer = x->buffer;
    int len = read(x->stdout[0], &buffer[x->buffer_pos], maxread);

    if (len > 0)
    {
        buffer[x->buffer_pos + len] = '\0';

        if (x->buffer_pos + len >= MAX_BUFFER)
            forceflush = TRUE;

        initng_plugin_readpipe(service, len, forceflush);
        x->buffer_pos = x->buffer_pos + len;
        if (forceflush)
            x->buffer_pos = 0;
    }
    else if (program_end)
    {
        /* No data, but notify plugins to flush whatever they haven't printed */
        initng_plugin_readpipe(service, 0, TRUE);
        x->buffer_pos = 0;
    }
    return len;
}

/* This does all the fancy pancy around the read */
void initng_process_readpipe(active_h * service, int program_end)
{
    int len;
    int fd_flags;
    process_h *x;

    /* Find out if we're starting or stopping a process */
    if (service->start_process)
    {
        x = service->start_process;
    }
    else if (service->stop_process)
    {
        x = service->stop_process;
    }
    else
        return;

    if (program_end)
    {
        /* Make sure last read doesn't block */
        fd_flags = fcntl(x->stdout[0], F_GETFL, 0);
        fcntl(x->stdout[0], F_SETFL, fd_flags | O_NONBLOCK);
    }

    len = initng_process_readpipe_read(service, x, program_end);

    if ((program_end) || len == -1)
    {
        /* Program is ending, close pipes */
        D_("Closing pipes for %s\n", service->name);
        if (x->stdout[0] > 0)
            close(x->stdout[0]);
        if (x->stdout[1] > 0)
            close(x->stdout[1]);
        x->stdout[0] = -1;
        x->stdout[0] = -1;
    }
}

/* this will poll all filedescriptors in g.fds[] db, for input, 
   and call right function, when event apere 
   returns FALSE if select() timed out. */
int initng_plugin_fd_poll(int timeout)
{
    fd_set readset;
    struct timeval tv;
    int i;
    int retval;
    int set = FALSE;
    active_h *tmp = NULL;

    FD_ZERO(&readset);

    /* First scan through active plugins that have listening file descriptors, and add them */
    for (i = 0; i < MAX_CALLERS; i++)
    {
        if (g.callers[i].ptype == FDWATCHERS
            && (g.callers[i].c.fdh.fds > 0
                || g.callers[i].c.fdh.call_module != NULL))
        {
            FD_SET(g.callers[i].c.fdh.fds, &readset);
            set = TRUE;
            D_("adding fd #%i to polls found on caller %i\n",
               g.callers[i].c.fdh.fds, i);
        }
    }

    /* Then add all file descriptors from process read pipes */
    while ((tmp = active_db_next(tmp)))
    {
        if ((tmp->start_process) && (tmp->start_process->stdout[0] > 0))
        {
            FD_SET(tmp->start_process->stdout[0], &readset);
            set = TRUE;
        }
        if ((tmp->stop_process) && (tmp->stop_process->stdout[0] > 0))
        {
            FD_SET(tmp->stop_process->stdout[0], &readset);
            set = TRUE;
        }
    }

    /* check if there are any set */
    if (!set)
    {
        D_("No filedescriptors\n");
        sleep(timeout);
        return (TRUE);
    }

    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    retval = select(256, &readset, NULL, NULL, &tv);
    if (errno == -1)
        printf("Select returned with %d %s\n", retval, strerror(errno));
    D_("select retval=%i\n", retval);

    if (retval == 0)
        return (FALSE);
    /* find matches, for calling right funtion to 
       fetch the incoming data */
    if (retval > 0)
    {
        D_("%d fd's active\n", retval);
        /* Scan through fdwatchers, check for matching fd's */
        for (i = 0; i < MAX_CALLERS; i++)
            if ((g.callers[i].ptype == FDWATCHERS
                 && (g.callers[i].c.fdh.fds > 0
                     || g.callers[i].c.fdh.call_module != NULL))
                && (FD_ISSET(g.callers[i].c.fdh.fds, &readset)))
            {
                D_("Calling pluginhandler for fd #%i\n",
                   g.callers[i].c.fdh.fds);
                (*g.callers[i].c.fdh.call_module) ();
                D_("pluginhandler for fd #%i returned!\n",
                   g.callers[i].c.fdh.fds);

                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
        /* Now scan through processes */
        tmp = NULL;
        while ((tmp = active_db_next(tmp)))
        {
            if ((tmp->start_process)
                && (FD_ISSET(tmp->start_process->stdout[0], &readset)))
            {
                initng_process_readpipe(tmp, FALSE);
                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
            if ((tmp->stop_process)
                && (FD_ISSET(tmp->stop_process->stdout[0], &readset)))
            {
                initng_process_readpipe(tmp, FALSE);
                /* Found match, that means we need to look for one less, if we've found all we should then return */
                retval--;
                if (retval == 0)
                    return (TRUE);
            }
        }
    }
    return (TRUE);
}

service_h *initng_modules_parse(char *name)
{
    int i;
    service_h *to_ret;

    for (i = 0; i < MAX_CALLERS; i++)
    {
        if (g.callers[i].ptype == PARSERS && g.callers[i].c.parser)
        {
            to_ret = (*g.callers[i].c.parser) (name);
            if (to_ret)
                return (to_ret);
        }
    }
    return (NULL);
}

/* called when a service status has changed. */
void initng_load_module_service_changed(active_h * service)
{
    int i;

    for (i = 0; i < MAX_CALLERS; i++)
        if (g.callers[i].ptype == WATCHERS && g.callers[i].c.watcher)
            (*g.callers[i].c.watcher) (service);
    return;
}

/* called when system state has changed. */
void initng_load_module_system_changed(h_sys_state state)
{
    int i;

    for (i = 0; i < MAX_CALLERS; i++)
        if (g.callers[i].ptype == SWATCHERS && g.callers[i].c.swatcher)
            (*g.callers[i].c.swatcher) (state);
    return;
}
