
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <sys/wait.h>                       /* waitpid() sa */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <stdio.h>                          /* printf() */

#include "initng_signal.h"
#include "initng_handler.h"
#include "initng_kill_handler.h"
#include "initng_global.h"
#include "main.h"

static void sigchild(int sig);
static void sigwinch(int sig);
static void sigint(int sig);
static void sigalarm(int sig);

volatile int initng_signal_got_keyboard_request = 0;
volatile int initng_signal_got_ctrl_alt_del = 0;

/* calld by signal SIGCHILD */
static void sigchild(int sig)
{
    int status = 0;             /* data got from waitpid, never used */
    pid_t killed;               /* pid of killed app */

    (void) sig;
    g.interupt = TRUE;
    D_("sigchild():  HHHHHHHHHHHHHHHHHHHHHHH   We lost a CHILD HHHHHHHHHHHHHHHHHHHHHHHHHHHH\n");
    while ((killed = waitpid(-1, &status, WNOHANG)) > 0)
    {
        D_("sigchild(): PID %i KILLED WITH:\n"
           "WEXITSTATUS %i\n"
           "WIFEXITED %i\n"
           "WIFSIGNALED %i\n"
           "WTERMSIG %i\n"
           "WCOREDUMP %i\n"
           "WIFSTOPPED %i\n"
           "WSTOPSIG %i\n"
           "\n",
           killed,
           WEXITSTATUS(status),
           WIFEXITED(status),
           WIFSIGNALED(status),
           WTERMSIG(status),
           WCOREDUMP(status), WIFSTOPPED(status), WSTOPSIG(status));
        handle_killed_by_pid(killed, WEXITSTATUS(status));
        g.interupt = TRUE;

        /* send this signal again, to make a loop, there might be new kills pending */
        /* IF SKIP WHILE LOOP raise(SIGCHLD); */
    }
}

/* calld by signal SIGWINCH */
static void sigwinch(int sig)
{
    (void) sig;
    g.interupt = TRUE;
    D_("sigwinch():     Got Keby Request\n");
    initng_signal_got_keyboard_request = 1;
}

/* sig alarm got */
static void sigalarm(int sig)
{
    (void) sig;
    term_stopping_services();
}

/* called by signal SIGINT */
static void sigint(int sig)
{
    (void) sig;
    g.interupt = TRUE;
    D_("sigint():     Got CTRL_ALT_DELETE\n");
    initng_signal_got_ctrl_alt_del = 1;
}

/* Hello this is main */
void enable_signals(void)
{
    struct sigaction sa;

    D_("enable_signals();\n");

    /* clear interrupt cache */
    /* TODO FIGUREITOUT - Catch signals */
    /*  signal(SIGPWR,sighandler); don't know what to do about it */
    /*  signal(SIGHUP,sighandler); ??? */
    /* SA_NOCLDSTOP -->  Get notification when childs stops living */
    /* SA_RESTART --> make certain system calls restartable across signals */
    /* What does this do in practise ?? */

    /* creat an signal set */
    sigemptyset(&sa.sa_mask);

    /* i dont get this one */
    sa.sa_sigaction = 0;


    /* SA_NOCLDSTOP = Dont give initng signal if we kill the app with SIGSTOP */
    /* SA_RESTART = call signal over again next time */
    sa.sa_flags = SA_RESTART | SA_NOCLDSTOP;

    sa.sa_handler = sigchild;
    sigaction(SIGCHLD, &sa, 0);             /* Dead children */

    sa.sa_handler = sigint;
    sigaction(SIGINT, &sa, 0);              /* ctrl-alt-del */

    sa.sa_handler = sigwinch;
    sigaction(SIGWINCH, &sa, 0);            /* keyboard request */

    sa.sa_handler = sigalarm;
    sigaction(SIGALRM, &sa, 0);             /* alarm, something has to be checked */
}

void disable_signals(void)
{
    struct sigaction sa;

    D_("disable_signals();\n");

    /* creat an signal set */
    sigemptyset(&sa.sa_mask);

    sa.sa_flags = SA_RESTART | SA_NOCLDSTOP;

    /* i dont get this one */
    sa.sa_sigaction = 0;
    sa.sa_handler = SIG_DFL;
    sigaction(SIGCHLD, &sa, 0);             /* Dead children */
    sigaction(SIGINT, &sa, 0);              /* ctrl-alt-del */
    sigaction(SIGWINCH, &sa, 0);            /* keyboard request */
    sigaction(SIGALRM, &sa, 0);             /* alarm, something has to be checked */

}
