
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdlib.h>                         /* malloc() free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <termios.h>
#include <stdio.h>
#include <sys/klog.h>
#include <dirent.h>
#include <sys/stat.h>
#include <fnmatch.h>

#include "initng_global.h"
#include "main.h"
#include "initng_signal.h"
#include "initng_history.h"
#include "initng_handler.h"
#include "initng_execute.h"                 /* new_environ() */
#include "active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_toolbox.h"
#include "initng_main.h"
#include "service_data_types.h"

/*
 * %%%%%%%%%%%%%%%%%%%%   main ()   %%%%%%%%%%%%%%%%%%%%
 */
#ifdef BUSYBOX
int init_main(int argc, char *argv[])
{
#else
int main(int argc, char *argv[])
{
#endif
    time_t last;                /* save the time here */

#ifdef DEBUG
    int loop_counter = 0;       /* a common used for counter */
#endif

    D_("main();\n");
    /* set console loglevel */
    klogctl(8, NULL, 1);

    /* get the time */
    last = time(NULL);

    /* initiate global variables */
    initng_new();

    if (getpid() == 1)
        s_set(I_AM_INIT);

    /* Load modules */
    {
        DIR *d;
        struct dirent *e;
        struct stat s;
        char *t = (char *) initng_calloc(100, sizeof(char));    /* allocate space for a filename */

        /* open plugin dir */
        if (!(d = opendir(INITNG_PLUGIN_DIR)))
        {
            F_("Unable top open plugin direcory " INITNG_PLUGIN_DIR ".\n");
            free(t);
            su_login();
        }

        /* get every entry */
        while ((e = readdir(d)) != NULL)
        {
            /* check for files, ending with .so.0.0 */
            if (fnmatch("*.so.0.0", e->d_name, 0) == 0)
            {
                /* set full path */
                strcpy(t, INITNG_PLUGIN_DIR "/");
                strcat(t, e->d_name);
                t[99] = '\0';               /* make sure there is no buffert overrun */
                /* stat the file, to check that it exits, and is a ordinary file */
                if (stat(t, &s) == 0 && S_ISREG(s.st_mode))
                {
                    D_("loading \"%s\"\n", t);
                    initng_load_module(t);  /* load the module */
                }
            }
        }
        closedir(d);
        free(t);

    }

    /* Pointer to starting arguments, this makes them GLOBAL */
    g.Argv = argv;
    g.Argc = argc;
    /* See if we are the first process born in the system */
    if (s_is(I_AM_INIT))
    {
        int fd;                 /* /dev/console */

        g.when_out = THEN_RESTART;
        reboot(RB_DISABLE_CAD);             /* Disable Ctrl + Alt + Delete */

        /* Try to open the console, but don't control it */
        if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) > 0)
        {
            D_("Opened /dev/console. Setting terminal options.\n");
            ioctl(fd, KDSIGACCEPT, SIGWINCH);   /* Accept signals from 'kbd' */
            close(fd);                      /* Like Ctrl + Alt + Delete signal? */
        }
        else
        {
            D_("Failed to open /dev/console. Setting options anyway.\n");
            ioctl(0, KDSIGACCEPT, SIGWINCH);    /* Accept signals from 'kbd' */
        }

        /* TODO: this block may be incorrect or incomplete */
        {
            struct termios tty;
            int fd;

            /*
             * TODO: /dev/console may still be open from before. Also, if it
             * fails to open, why try to finish the block?
             */
            if ((fd = open("/dev/console", O_RDWR | O_NOCTTY)) < 0)
            {
                F_("main(): can't open /dev/console.\n");
            }

            (void) tcgetattr(fd, &tty);

            tty.c_cflag &= CBAUD | CBAUDEX | CSIZE | CSTOPB | PARENB | PARODD;
            tty.c_cflag |= HUPCL | CLOCAL | CREAD;

            tty.c_cc[VINTR] = 3;            /* ctrl('c') */
            tty.c_cc[VQUIT] = 28;           /* ctrl('\\') */
            tty.c_cc[VERASE] = 127;
            tty.c_cc[VKILL] = 24;           /* ctrl('x') */
            tty.c_cc[VEOF] = 4;             /* ctrl('d') */
            tty.c_cc[VTIME] = 0;
            tty.c_cc[VMIN] = 1;
            tty.c_cc[VSTART] = 17;          /* ctrl('q') */
            tty.c_cc[VSTOP] = 19;           /* ctrl('s') */
            tty.c_cc[VSUSP] = 26;           /* ctrl('z') */

            /*
             *      Set pre and post processing
             */
            tty.c_iflag = IGNPAR | ICRNL | IXON | IXANY;
            tty.c_oflag = OPOST | ONLCR;
            tty.c_lflag = ISIG | ICANON | ECHO | ECHOCTL | ECHOPRT | ECHOKE;

            /*
             *      Now set the terminal line.
             *      We don't care about non-transmitted output data
             *      and non-read input data.
             */
            (void) tcsetattr(fd, TCSANOW, &tty);
            (void) tcflush(fd, TCIOFLUSH);
            (void) close(fd);
        }

    }

    /* Configure signal handlers */
    (void) enable_signals();


    /* If no starting services found, use default */
    if (!s_get_string(S_RUNLEVEL))
    {
        if (s_is(I_AM_INIT))
        {
            s_set_string(S_RUNLEVEL, i_strdup("default"));
        }
        else
        {
            s_set_string(S_RUNLEVEL, i_strdup("fake-default"));
        }
    }

    start_new_service_named(s_get_string(S_RUNLEVEL));
    start_initiating_service();             /* Start the first service */

    /* %%%%%%%%%%%%%%%   MAIN LOOP   %%%%%%%%%%%%%%% */
    for (;;)
    {
        time_t now;             /* Store the start time of each loop */

        D_("main():mainloop: %i\n", loop_counter++);

        /* Update current time */
        now = time(NULL);

        /* Check some global variables, set by sigactions */

        /* If we got an interrupt load ctrlaltdel */
        if (initng_signal_got_ctrl_alt_del)
        {
            initng_signal_got_ctrl_alt_del = 0;
            start_new_service_named("ctrlaltdel");
        }

        /* If we got a keybord request load kbreq */
        if (initng_signal_got_keyboard_request)
        {
            initng_signal_got_keyboard_request = 0;
            start_new_service_named("kbreq");
        }

        /*
         * If now is before last, the time was set backwards.
         * If so, compensate the change in the whole process_db
         */
        if (now < last)
        {
            D_("main():mainloop: The clock was reset. Compensating...\n");
            active_db_compensate_time(last - now);
            history_db_compensate_time(last - now);
        }

        /*
         * If we moved more than an hour ahead, work backwards in one hour
         * chunks until we are within one hour of the original.
         *
         * TODO: why is this important?
         */
        while (now - last >= 3600)
        {
            active_db_compensate_time(3600);
            history_db_compensate_time(3600);
            last = last + 3600;
        }
        last = now;                         /* Remember last time */

        /* 
         * Start launching dependencies, until no more services are marked.
         * The dependencies db should continue to stabilize through each run.
         */
        while (g.interupt)
        {
            g.interupt = FALSE;
            try_start_wait_deps();
            D_("main():miniloop (will continue as long as interrupt is set)\n");
        }
        D_("main(): Finished loading dependencies!\n");

        /* Look for any processes that won't DIE */

        /* term_stopping_services(); */


        /*
         * Check if there are any running processes left, otherwise quit
         *
         * TODO: it isn't clear that ready_to_quit() is what stops initng.
         * "ready_to_quit" sounds like a condition test, we should respond to
         * it here with a handler.
         *
         * if( (status = ready_to_quit()) )
         * {
         *     status_change_handler(status);
         * }
         */
        ready_to_quit();

        /* call hooks that want to be run every main loop */
        initng_plugin_main();
        /*
         * See if anything is happening on the control channel
         *  - wait for 60seconds
         */
        if (!g.interupt)
        {
            if (g.sys_state == STATE_STOPING)
            {
                if (!initng_plugin_fd_poll(5000))
                {
                    F_("Waited for 5 minutes when shutting down, setting verbose mode!\n");
                    g.verbose = TRUE;
                }
            }
            else
                initng_plugin_fd_poll(60000);
            /*remote_control_poll(60000); */
        }

    }                                       /* End main loop */
}
