
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <errno.h>

#include "initng_global.h"
#include "initng_toolbox.h"
#include "open_read_close.h"

void bailout(int *, char **buffert);


int open_read_close(char *filename, char **buffert)
{
    int conf_file;              /* File descriptor for configfile */
    off_t fsize;                /* (Temp) file size holder */
    size_t filesize;            /* Remember filesize here */
    int res;                    /* Result of read */

    /* Mark *buffer and conf_file as not set, for cleanup if bailing out... */

    *buffert = NULL;
    conf_file = -1;

    /* */

    conf_file = open(filename, O_RDONLY);   /* Open configfile. */

    if (conf_file == -1)
    {
        D_("open_read_close(%s) CANT OPEN FILE!\n", filename);

        bailout(&conf_file, buffert);
        return (FALSE);
    }

    fsize = lseek(conf_file, 0, SEEK_END);  /* Get filesize. */

    if (fsize == (off_t) (-1))
    {
        F_("initng_parser_open_read_close(%s): %s\n",
           filename, strerror(errno));

        bailout(&conf_file, buffert);
        return (FALSE);
    }

    if (fsize >= MAX_FILESIZE)
    {
        F_("initng_parser_open_read_close(%s) Config file to large\n"
           "Change source to be able to allocate bigger files\n"
           "max size is %i, your is %ld and %ld to big.\n\n",
           filename, MAX_FILESIZE, fsize, (fsize - MAX_FILESIZE));

        bailout(&conf_file, buffert);
        return (FALSE);
    }

    /* Go back to start. */

    if (lseek(conf_file, 0, SEEK_SET) == (off_t) (-1))
    {
        F_("initng_parser_open_read_close(%s): %s\n",
           filename, strerror(errno));

        bailout(&conf_file, buffert);
        return (FALSE);
    }

    /* Fit in size_t */

    filesize = (size_t) fsize;

    /* Allocate a file buffert */

    *buffert = initng_calloc((size_t) 1, (filesize + 1));

    if (*buffert == NULL)
    {
        F_("initng_parser_open_read_close(%s): calloc %d failed\n",
           filename, (int) filesize + 1);

        bailout(&conf_file, buffert);
        return (FALSE);
    }

    /* Read whole file */

    res = read(conf_file, *buffert, filesize);

    if (res == -1 || (size_t) res != filesize)
    {
        F_("initng_parser_open_read_close(%s): Failed to read %d bytes\n",
           filename, (int) filesize);

        bailout(&conf_file, buffert);
        return (FALSE);
    }

    /* Normally we wouldn't care about this, but as this is init(ng)? */

    if (close(conf_file) < 0)
    {
        F_("initng_parser_open_read_close(%s): %s\n",
           filename, strerror(errno));

        bailout(&conf_file, buffert);
        return (FALSE);
    }

    (*buffert)[filesize] = '\0';            /* Null terminate *buffer */

    return (TRUE);
}

/* Avoid using goto sending a pointer to conf_file */
void bailout(int *p_conf_file, char **buffert)
{
    /* if conf_file != -1 it is open */

    if (*p_conf_file != -1)
        (void) close(*p_conf_file);         /* Ignore result this time */

    *p_conf_file = -1;

    /* *buffert != NULL => we have called calloc, so free it */

    if (*buffert)
        free(*buffert);

    *buffert = NULL;
}
