
/* Initng, a next generatind sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#include <string.h>
#include <stdlib.h>

#include "service_data_types.h"
#include "initng_global.h"
#include "initng_toolbox.h"


e_sdt initng_sdt_add(e_dt d_type, char *name, e_sdt force_id)
{
    int i = 0;

    D_("try add %s\n", name);
    if (g.option_table == NULL)
    {
        g.option_table =
            (s_data_row *) initng_calloc(MAX_DATA_TYPES, sizeof(s_data_row));
        if (!g.option_table)
            return (-1);
    }

    /* it might allredy been added */
    if ((i = initng_sdt_find(name, d_type)) > 0)
        return (i);

    /* fast forward to a free spot */
    i = 0;
    while (i < MAX_DATA_TYPES && g.option_table[i].opt_type
           && g.option_table[i].opt_name)
        i++;
    D_("free spot att %i to add service on\n", i);

    if (i == MAX_DATA_TYPES)
    {
        D_("Reached MAX_DATA_TYPES!\n");
        return (-1);
    }

    g.option_table[i].opt_type = d_type;
    g.option_table[i].opt_name = name;
    g.option_table[i].chars = strlen(name);

    if (force_id > 0)
    {
        g.option_table[i].opt_id = force_id;
        D_("force added %s type %i id %i\n", name, d_type, force_id);
        return (force_id);
    }

    /* search the db for a opt_id that is unused */
    i = 1;
    while (i < MAX_DATA_TYPES && g.option_table[i].opt_id > 0)
        i++;

    g.option_table[i].opt_id = i;
    D_("dynamic added \"%s\" type %i id %i\n", name, d_type, i);
    return (i);
}

void initng_sdt_del(e_sdt id)
{
    int i = 0;

    while (i < MAX_DATA_TYPES)
        if (g.option_table[i].opt_id == id)
        {
            g.option_table[i].opt_id = 0;
            g.option_table[i].opt_type = U_D_T;
            if (g.option_table[i].opt_name)
            {
                free(g.option_table[i].opt_name);
                g.option_table[i].opt_name = NULL;
            }
        }
}

void initng_sdt_free_all(void)
{
    int i = 0;

    D_("TODO, make a way to detect if opt_name is a constant or not.\n");
    return;
    for (; i < MAX_DATA_TYPES; i++)
    {
        if (g.option_table[i].opt_id && g.option_table[i].opt_type != U_D_T)
        {
            g.option_table[i].opt_id = 0;
            g.option_table[i].opt_type = U_D_T;
            if (g.option_table[i].opt_name)
            {
                free(g.option_table[i].opt_name);
                g.option_table[i].opt_name = NULL;
            }
            g.option_table[i].chars = 0;
        }
    }
}

e_sdt initng_sdt_find(char *name, e_dt type)
{
    int i = 0;
    int chars = strlen(name);

    D_("looking for %s.\n", name);
    for (; i < MAX_DATA_TYPES; i++)
        if ((type == U_D_T || g.option_table[i].opt_type == type) &&
            chars == g.option_table[i].chars &&
            strcmp(g.option_table[i].opt_name, name) == 0)
        {
            D_("Found %s in %i\n", name, i);
            return (g.option_table[i].opt_id);
        }
    D_("Did not found %s\n", name);
    return (0);
}

void initng_sdt_add_defaults(void)
{
    initng_sdt_add(STRING, "name", NAME);
    initng_sdt_add(SET, "respawn", RESPAWN);
    initng_sdt_add(UNSET, "norespawn", RESPAWN);
    initng_sdt_add(SET, "up_when_done", UP_WHEN_DONE);
    initng_sdt_add(SET, "up_when_exit", UP_WHEN_DONE);
    initng_sdt_add(SET, "up_when_is_pid_set", UP_WHEN_PID_SET);
    initng_sdt_add(SET, "up_when_pid_set", UP_WHEN_PID_SET);
    initng_sdt_add(SET, "up_when_started", UP_WHEN_STARTED);
    initng_sdt_add(SET, "up_when_service_is_started", UP_WHEN_STARTED);
    initng_sdt_add(SET, "up_when_service_started", UP_WHEN_STARTED);
    initng_sdt_add(STRINGS, "daemon", DAEMON);
    initng_sdt_add(STRING, "daemon_args", DAEMON_ARGS);
    initng_sdt_add(STRING, "start_args", START_ARGS);
    initng_sdt_add(SCRIPT, "start {", START_SCRIPT);
    initng_sdt_add(SCRIPT, "start{", START_SCRIPT);
    initng_sdt_add(SCRIPT, "start_script", START_SCRIPT);
    initng_sdt_add(SCRIPT, "start_bash", START_SCRIPT);
    initng_sdt_add(STRINGS, "start", START);
    initng_sdt_add(STRING, "stop_args", STOP_ARGS);
    initng_sdt_add(SCRIPT, "stop {", STOP_SCRIPT);
    initng_sdt_add(SCRIPT, "stop{", STOP_SCRIPT);
    initng_sdt_add(SCRIPT, "stop_script", STOP_SCRIPT);
    initng_sdt_add(SCRIPT, "stop_bash", STOP_SCRIPT);
    initng_sdt_add(STRINGS, "stop", STOP);
    initng_sdt_add(STRING, "pid_file", PID_FILE);
    initng_sdt_add(STRING, "pidfile", PID_FILE);
    initng_sdt_add(STRINGS, "needs", NEED);
    initng_sdt_add(STRINGS, "need", NEED);
    initng_sdt_add(STRINGS, "depends", NEED);
    initng_sdt_add(STRINGS, "depend", NEED);
    initng_sdt_add(STRINGS, "requires", NEED);
    initng_sdt_add(STRINGS, "require", NEED);
    initng_sdt_add(STRINGS, "before", BEFORE);
    initng_sdt_add(STRINGS, "uses", USE);
    initng_sdt_add(STRINGS, "use", USE);
    initng_sdt_add(STRINGS, "after", USE);
    initng_sdt_add(STRINGS, "provides", PROVIDE);
    initng_sdt_add(STRINGS, "provide", PROVIDE);
    initng_sdt_add(STRINGS, "rdepends", RDEPENDS);
    initng_sdt_add(STRING, "from_file", FROM_FILE);
    initng_sdt_add(STRINGS, "environ", ENV);
    initng_sdt_add(STRINGS, "env", ENV);
    initng_sdt_add(INT, "nice", NICE);
    initng_sdt_add(INT, "suid", SUID);
    initng_sdt_add(INT, "user", SUID);
    initng_sdt_add(INT, "set_user", SUID);
    initng_sdt_add(INT, "set_uid", SUID);
    initng_sdt_add(INT, "setuid", SUID);
    initng_sdt_add(INT, "chuid", SUID);
    initng_sdt_add(INT, "udelay", UDELAY);
    initng_sdt_add(INT, "delay", DELAY);
    initng_sdt_add(INT, "upause", UDELAY);
    initng_sdt_add(INT, "pause", DELAY);
    initng_sdt_add(STRING, "chdir", CHDIR);
    initng_sdt_add(STRING, "chroot", CHROOT);
    initng_sdt_add(STRINGS, "also_stop", ALSO_STOP);
    initng_sdt_add(STRING, "also_start", ALSO_START);
    initng_sdt_add(SET, "done_when_faild", DONE_WHEN_FAIL);
    initng_sdt_add(SET, "done_when_fail", DONE_WHEN_FAIL);
    initng_sdt_add(SET, "done_on_fail", DONE_WHEN_FAIL);
}
