
/* Initng, a 6next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"

#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>
#include "process_db.h"
#include "service_db.h"
#include "initng_global.h"
#include "string_tools.h"
#include "active_db.h"
#include "initng_toolbox.h"

#include "service_data_types.h"
/*service_h **config_database = NULL; */

/*static int next_service_to_alloc = 0; */

/* change service_h pointer in database, maybe on a realloc */
void service_db_change(service_h * from, service_h * to)
{
    int i = 0;

    assert(from);
    for (i = 0; i < g.config_allocated; i++)
        if (g.config_database[i] == from)
            g.config_database[i] = to;
}

service_h *service_db_get_next(service_h * current)
{
    int i = 0;

    assert(g.config_database);

    if (current)
        for (; i < g.config_allocated; i++)
            if (g.config_database[i] == current)
            {
                i++;
                break;
            }

    for (; i < g.config_allocated; i++)
        if (g.config_database[i])
            return (g.config_database[i]);
    return (NULL);
}

/* add service to data structure, return index or -1 */
service_h *service_db_add(service_h * s)
{
    service_h *tmp_serv = 0;
    int i = 0;

    assert(s);
    assert(g.config_database);

    /* Check that we are unique */
    if ((tmp_serv = service_db_find_by_name(s->name)))
    {
        D_("service_db_add(%s): WARNING: DUPLICATE. Already added!\n",
           s->name);
        return (tmp_serv);
    }

    /* find a free spot */
    while (i < g.config_allocated)
    {
        if (!g.config_database[i])
        {

            g.config_database[i] = s;

            D_("service_db_add(%s): added at entry g.config_database[%i].\n",
               g.config_database[i]->name, i);
            return (g.config_database[i]);
        }
        i++;
    }
    /* if not found, we have to make config_database larger */



    return (NULL);
}

/* returns a service mallocated copy from s */
int service_db_copy(service_h ** new_serv, char *name, service_h * s)
{
    int i;

    assert(s);
    assert(name);
    /* allocate new service */
    *new_serv = (service_h *) initng_calloc(1, sizeof(service_h));
    /* copy service structure */
    memcpy(*new_serv, s, sizeof(service_h));
    /* name has to diff */
    (*new_serv)->name = name;

    /* walk through all data, and copy all strings. to avoid unwanted surprises. */
    for (i = 0; i < M_D_F; i++)
    {
        if ((*new_serv)->data[i].type != EMPTY && (*new_serv)->data[i].t.s)
        {
            (*new_serv)->data[i].t.s = i_strdup((*new_serv)->data[i].t.s);
        }
    }

    D_("copy_service(%s);\n", s->name);

    return (TRUE);
}

/* returns an empty service mallocated with default values */
int service_db_new(service_h ** d, char *name, e_type type)
{
    int i;

    assert(name);
    D_("default_service();\n");
    *d = (service_h *) initng_calloc(1, sizeof(service_h));
    if (!*d)
    {
        F_("OUT OF MEMORY default_service().\n");
        return (FALSE);
    }
    (*d)->name = name;
    (*d)->type = type;
    for (i = 0; i < M_D_F; i++)
    {
        (*d)->data[i].type = EMPTY;
        (*d)->data[i].t.s = NULL;
    }
    return (TRUE);
}

/* find service by name, returning pointer to service */
service_h *service_db_find_by_name(char *find)
{
    int i;

    assert(find);
    assert(g.config_database);

    for (i = 0; i < g.config_allocated; i++)
    {
        if (g.config_database[i])
        {
            if (st_service_cmp(find, g.config_database[i]->name))
            {
                D_("service_db_find_by_name(%s): Found g.config_database[%i].\n", find, i);
                return (g.config_database[i]);
            }
        }
    }

    return (NULL);
}

/* free all the dynamically allocated strings in this service */
int service_db_free(service_h * to_free)
{
    assert(to_free);
    D_("service_db_free(%s);\n", to_free->name);

    /* change other database relations to this service to null */
    service_db_change(to_free, NULL);
    active_db_change_service_h(to_free, NULL);
    /* free all strings in service */
    service_db_remove_all(to_free);

    /* free service name */
    if (to_free->name)
        free(to_free->name);
    free(to_free);
    return (TRUE);
}

/* Walk through every service and clear it */
int service_db_init(void)
{
    /* allocate config database */
    g.config_database =
        (service_h **) initng_calloc(MAX_SERVICE + 1, sizeof(service_h *));
    g.config_allocated = MAX_SERVICE;
    if (!g.config_database)
    {
        F_("Out of memory, can't allocate service database!\n");
        return (FALSE);
    }
    D_("service_db_init();\n");
    return (TRUE);
}

/* Walk through every service and clear it */
void service_db_clear_all(void)
{
    int i;

    assert(g.config_database);
    D_("service_db_clear();\n");
    for (i = 0; i < g.config_allocated; i++)
    {
        if (g.config_database[i])
        {
            service_db_free(g.config_database[i]);
            g.config_database[i] = NULL;
        }
    }
}

/* Walk through every service and clear it */
void service_db_free_all(void)
{
    assert(g.config_database);
    service_db_clear_all();
    free(g.config_database);
    g.config_database = NULL;
}

char *service_db_get_string(e_sdt type, service_h * from_service)
{
    assert(from_service);
    if (type == NAME)
        return (from_service->name);

    return (d_get_string(type, from_service->data));
}


/* this has to be only one of */
void service_db_set_string(e_sdt type, service_h * from_service, char *string)
{
    assert(from_service);
    assert(string);

    if (type == NAME)
    {
        from_service->name = string;
        return;
    }
    d_set_string(type, from_service->data, string);
}


int service_db_is(e_sdt type, service_h * from_service)
{
    assert(from_service);
    if (type == NAME)
    {
        if (from_service->name)
            return (TRUE);
        else
            return (FALSE);
    }
    return (d_is(type, from_service->data));
}
