
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include "initng.h"
#include "initng_global.h"

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "struct_data.h"


char *d_get_string(e_sdt type, s_data * d)
{
    int i = 0;

    assert(d);

    /* search for type */
    for (; i < M_D_F; i++)
        if (d[i].type == type && d[i].t.s)
            return (d[i].t.s);

    /* did not get any match */
    D_("String type %i doesn't exist.\n", type);
    return (NULL);
}

int d_get_int(e_sdt type, s_data * d)
{
    int i = 0;

    assert(d);
    /* search for type */
    for (; i < M_D_F; i++)
        if (d[i].type == type && d[i].t.i)
            return (d[i].t.i);

    /* did not get any match */
    D_("Int value, type %i doesn't exist in db.\n", type);
    return (0);
}

char *d_get_next_string(e_sdt type, s_data * d, char *last)
{
    int i = 0;

    assert(d);

    /* fast forward */
    if (last != NULL)
        for (i = 0; i < M_D_F; i++)
            if (d[i].type == type && d[i].t.s == last)
            {
                i++;
                break;
            }

    /* search for type */
    for (; i < M_D_F; i++)
        if (d[i].type == type && d[i].t.s)
            return (d[i].t.s);

    /* did not get any match */
    D_("String type %i doesn't exist in db.\n", type);
    return (NULL);
}

/* this is a function to add a string, that it can exist many of */
void d_set_another_string(e_sdt type, s_data * d, char *string)
{
    int i;

    assert(d);
    assert(string);

    /* find a free spot */
    for (i = 0; i < M_D_F; i++)
        if (d[i].type == EMPTY)
        {
            /* add type to database */
            d[i].type = type;
            d[i].dt = STRINGS;
            d[i].t.s = string;
            return;
        }
}

/* this has to be only one of */
void d_set_string(e_sdt type, s_data * d, char *string)
{
    int i;

    assert(d);
    assert(string);

    /* check if it is enabled already */
    for (i = 0; i < M_D_F; i++)
    {
        if (d[i].type == type && d[i].dt == STRING)
        {
            if (d[i].t.s)
                free(d[i].t.s);
            d[i].t.s = string;
            return;
        }
    }

    /* find a free spot */
    for (i = 0; i < M_D_F; i++)
        if (d[i].type == EMPTY)
        {
            /* add type to database */
            d[i].type = type;
            d[i].dt = STRING;
            d[i].t.s = string;
            return;
        }
}

/* this has to be only one of */
void d_set_int(e_sdt type, s_data * d, int value)
{
    int i;

    assert(d);

    /* check if it is enabled already */
    for (i = 0; i < M_D_F; i++)
    {
        if (d[i].type == type && d[i].dt == INT)

        {
            d[i].t.i = value;
            return;
        }
    }

    /* find a free spot */
    for (i = 0; i < M_D_F; i++)
        if (d[i].type == EMPTY)
        {
            /* add type to database */
            d[i].type = type;
            d[i].dt = INT;
            d[i].t.i = value;
            return;
        }
}


/* Walk through and count */
int d_count_type(e_sdt type, s_data * d)
{
    int i;
    int count = 0;

    assert(d);
    if (type <= 0)
    {
        for (i = 0; i < M_D_F; i++)
            if (d[i].type != EMPTY)
                count++;
        return count;
    }

    /* check if it is enabled already */
    for (i = 0; i < M_D_F; i++)
        if (d[i].type == type)
            count++;
    return count;
}

int d_is(e_sdt type, s_data * d)
{
    int i;

    assert(d);

    /* search for type */
    for (i = 0; i < M_D_F; i++)
    {
        if (d[i].type == type)
        {
            return (TRUE);
        }
    }
    /* did not get any match */
    return (FALSE);
}

void d_set(e_sdt type, s_data * d)
{
    int i;

    assert(d);

    /* check if it is enabled already */
    if (d_is(type, d))
        return;

    /* add type to database */
    for (i = 0; i < M_D_F; i++)
        if (d[i].type == EMPTY)
        {
            d[i].type = type;
            d[i].dt = SET;
            d[i].t.s = NULL;
            return;
        }
}

void d_remove(e_sdt type, s_data * d)
{
    int i;

    assert(d);
    /* check if it is enabled already */
    for (i = 0; i < M_D_F; i++)
        if (d[i].type == type)
        {
            if (d[i].t.s)
                free(d[i].t.s);
            d[i].type = EMPTY;
        }
}

/* walk through and kill all strings */
void d_remove_all(s_data * d)
{
    int i;

    assert(d);
    /* check if it is enabled already */
    for (i = 0; i < M_D_F; i++)
    {
        if (d[i].type != EMPTY &&
            (d[i].dt == STRING || d[i].dt == STRINGS) && d[i].t.s)
            free(d[i].t.s);
        d[i].type = EMPTY;
    }
}
